from typing import cast

import pandas as pd
import numpy as np
import logging

from ..domain.mappings import (
    COUNTRY_TO_IRENA_MAP,
    COUNTRY_TO_REGION_MAP,
    COUNTRY_TO_CODE_MAP,
    COUNTRY_TO_SSP_REGION_MAP,
)

from steelo.utilities.global_variables import LEARNING_RATES, LIFETIME, YEARLY_DETERIORATION_RATE
from steelo.config import settings

# ----------------------------------------------------- FUNCTION DEF -----------------------------------------------------


def load_capacity_data(base_yr: int) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Loads the capacity factor and capacity projections data for solar PV and onshore wind.

        - Capacity factors: daily variable capacity factors (i.e., ratio between the actual and the theoretical maximum
        output when running always at full capacity; normal CF * 24) and size of bare land areas (i.e., number of km^2
        available to build on). Temporal res.: yearly. Spatial resolution: country level. Units: MW/km^2. Assumptions:
        capacity factors derived under the constraint of only using bare land areas. Source: Systemiq internal.

        - Historical capacity: historically installed capacity data for solar PV and onshore wind. Temporal res.: yearly.
        Time: 2023. Spatial res.: country and regional (244 countries). Units: MW.
        Source: IRENA (https://pxweb.irena.org/pxweb/en/IRENASTAT/IRENASTAT__Power%20Capacity%20and%20Generation/Country_ELECSTAT_2024_H2.px/).

        - Capacity projections (SSP-RCP): capacity buildout projections from 2005 to 2100. Temporal res.: decadal.
        Spatial res.: regional (5 regions). Units: GW. Source: IIASA (https://tntcat.iiasa.ac.at/SspDb/dsd?Action=htmlpage&page=welcome).
        Scenario selection:
            - SSP1-2.6: Sustainability; low emissions in line with 2°C target; low adaptation and mitigation challenges
            - SSP2-4.5: Middle of the Road; moderate emissions; medium adaptation and mitigation challenges
            - SSP3-baseline: Regional Rivalry; high emissions from no global climate policy (BAU); high adaptation and mitigation challenges
    """
    logging.info("Loading capacity data")
    # Load capacity factors data
    capacity_fact_file_path = settings.project_root / "data" / "capacity_factors.csv"
    capacity_factors_data = pd.read_csv(capacity_fact_file_path, index_col=0)

    # Load historical capacity data
    historical_capacity_file_path = settings.project_root / "data" / f"installed_renewables_capacity_{str(base_yr)}.csv"
    historical_capacity_data = pd.read_csv(historical_capacity_file_path, header=0)

    # Load capacity projections data
    solar_capacity_proj_file_path = settings.project_root / "data" / "ssp_rcp_solar_capacity_projections.xlsx"
    wind_capacity_proj_file_path = settings.project_root / "data" / "ssp_rcp_onshore_wind_capacity_projections.xlsx"
    solar_capacity_proj_data = pd.read_excel(solar_capacity_proj_file_path, usecols="A:P", header=0, skipfooter=2)
    wind_capacity_proj_data = pd.read_excel(wind_capacity_proj_file_path, usecols="A:P", header=0, skipfooter=2)

    return capacity_factors_data, historical_capacity_data, solar_capacity_proj_data, wind_capacity_proj_data


def load_cost_data(base_yr: int) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Loads the CAPEX data for solar and onshore wind, and other cost data.
        - CAPEX: Total cost, cost of hardware, and cost of non-hardware. Spatial resolution: regional level (16 regions
        worldwide). Year: 2022. Units: USD/kW. Source: Systemiq internal.
        - Other costs: Output from economic model. Loaded for cost of capital.
    """
    logging.info("Loading cost data")
    # Load solar CAPEX data
    solar_capex_file_path = settings.project_root / "data" / f"capex_solar_{str(base_yr)}.csv"
    solar_capex_data = pd.read_csv(solar_capex_file_path, index_col=0)

    # Load onshore wind CAPEX data
    wind_capex_file_path = settings.project_root / "data" / f"capex_onshore_wind_{str(base_yr)}.csv"
    wind_capex_data = pd.read_csv(wind_capex_file_path, index_col=0)

    # Load other costs data
    cost_of_x_path = settings.project_root / "data" / "cost_of_x.json"
    cost_data = pd.read_json(cost_of_x_path)

    return solar_capex_data, wind_capex_data, cost_data


def prepare_capex_data_per_tech(raw_capex_data: pd.DataFrame, tech_name: str) -> pd.DataFrame:
    """
    Renames the columns to include the technology name, resets the index, and drops irrelevant columns.
    """

    clean_capex_data = raw_capex_data.copy()
    clean_capex_data = clean_capex_data.reset_index().rename(
        columns={
            "Total capex (USD/kW)": f"Total capex {tech_name} (USD/kW)",
            "Hardware capex (USD/kW)": f"Hardware capex {tech_name} (USD/kW)",
            "Non-hardware capex (USD/kW)": f"Non-hardware capex {tech_name} (USD/kW)",
            "Opex (%)": f"Opex {tech_name} (%)",
        }
    )
    clean_capex_data.drop(columns=["Source"], inplace=True)

    return clean_capex_data


def preprocess_cost_data(
    solar_capex: pd.DataFrame,
    wind_capex: pd.DataFrame,
    cost_data: pd.DataFrame,
    countries_df: pd.DataFrame,
    country_to_region_map: dict,
    country_to_code_map: dict,
) -> pd.DataFrame:
    """
    Preprocesses the CAPEX and cost of capital data for solar and wind technologies. Maps the regional CAPEX data to countries,
    fills missing cost of capital values (about 20%) with the global average, and merges all data into a single df.
    """
    logging.info("Preprocessing cost data")
    # CAPEX
    solar_capex_pp = prepare_capex_data_per_tech(solar_capex, "solar")
    wind_capex_pp = prepare_capex_data_per_tech(wind_capex, "wind")
    renewable_capex = pd.merge(solar_capex_pp, wind_capex_pp, on="Region", how="outer")
    countries_ = countries_df.copy()
    countries_["Region"] = countries_["Country"].map(country_to_region_map)
    renewable_capex_per_country = pd.merge(countries_, renewable_capex, on="Region", how="left")

    # Cost of capital
    cost_of_capital = cost_data.copy()[["Country code", "Cost of capital - renewables"]].rename(
        columns={"Cost of capital - renewables": "Cost of capital (%)"}
    )
    country_to_code_map_r = {v: k for k, v in country_to_code_map.items()}
    cost_of_capital["Country"] = cost_of_capital["Country code"].map(country_to_code_map_r)
    cost_of_capital.drop(columns=["Country code"], inplace=True)

    # All costs
    cost_per_country = (
        pd.merge(renewable_capex_per_country, cost_of_capital, on="Country", how="left")
        .set_index("Country")
        .sort_index()
    )
    cost_per_country.drop(columns=["Region"], inplace=True)
    cost_per_country.fillna(cost_per_country.select_dtypes(include=[np.number]).mean(), inplace=True)

    return cost_per_country


def preprocess_capacity_factors(cap_fact: pd.DataFrame) -> pd.DataFrame:
    """
    Restructures the data, divides all data by 24 to get hourly capacity factors instead of daily ones, and
    fills missing values with the global average.
    """
    logging.info("Preprocessing capacity factors data")
    cap_fact_ = (
        cap_fact.transpose()
        .reset_index()
        .rename(
            columns={
                "index": "Country",
                "pvDaily_total": "Capacity factor solar (%)",
                "windDaily_total": "Capacity factor wind (%)",
            }
        )
    )
    for tech in ["solar", "wind"]:
        cap_fact_[f"Capacity factor {tech} (%)"] = cap_fact_[f"Capacity factor {tech} (%)"] / 24
    cap_fact_.drop(columns=["solar_area", "onshore_wind_area"], inplace=True)
    cap_fact_.fillna(cap_fact_.select_dtypes(include=[np.number]).mean(), inplace=True)
    cap_fact_ = cap_fact_.set_index("Country")

    return cap_fact_


def preprocess_historical_renewable_capacity(
    hist_cap: pd.DataFrame, countries: pd.DataFrame, country_to_irena_map: dict
) -> tuple[pd.Series, pd.Series]:
    """
    Maps country names from different data sources, fills missing values with the global average, and splits the data into solar and wind.
    """
    logging.info("Preprocessing historical capacity data")
    hist_cap_ = hist_cap.copy().rename(columns={"Country": "IRENA Country"})
    countries_ = countries.copy()
    countries_["IRENA Country"] = countries_["Country"].map(country_to_irena_map)
    cap_now = (
        pd.merge(countries_, hist_cap_, on="IRENA Country", how="left")
        .set_index("Country")
        .sort_index()
        .drop(columns=["IRENA Country"])
    )
    for tech in ["solar", "wind"]:
        cap_now[f"Capacity {tech} (MW)"] = pd.to_numeric(cap_now[f"Capacity {tech} (MW)"], errors="coerce")
        cap_now[f"Capacity {tech} (MW)"] = cap_now[f"Capacity {tech} (MW)"].fillna(
            cap_now[f"Capacity {tech} (MW)"].mean()
        )
    solar_cap_now = cap_now["Capacity solar (MW)"]
    wind_cap_now = cap_now["Capacity wind (MW)"]

    return solar_cap_now, wind_cap_now


def correct_ssp_projections_with_current_capacity(
    current_capacity: pd.Series | pd.DataFrame,
    raw_capacity_data: pd.DataFrame,
    countries_df: pd.DataFrame,
    country_to_ssp_region_map: dict,
    base_yr: int,
    tech: str,
) -> pd.DataFrame:
    """
    Corrects the magnitude of the SSP projections with the current capacity data. Only the gradient of the SSP projections is used to project
    these current data until 2100. Expands the decadal SSP projections to yearly resolution via linear interpolation.
    """
    logging.info(f"Correcting SSP projections with current capacity data for {tech}")
    countries_ = countries_df.copy()
    countries_["Region"] = countries_["Country"].map(country_to_ssp_region_map)
    capacity_per_country = pd.merge(raw_capacity_data, countries_, on="Region", how="left")

    # Fill missing values with global average
    capacity_per_country.fillna(capacity_per_country.select_dtypes(include=[np.number]).mean(), inplace=True)

    # Set index and drop irrelevant columns
    capacity_per_country = capacity_per_country.set_index(["Country", "Scenario"])
    capacity_per_country.drop(columns=["Model", "Variable", "Region", "Unit", 2005, 2010], inplace=True)
    year_cols = cast(
        list[int], [col for col in capacity_per_country.columns if isinstance(col, int) and 1900 <= int(col) <= 3000]
    )  # Select only years
    capacity_per_country = capacity_per_country.loc[:, year_cols]

    # Expand decadal data to yearly and interpolate linearly
    capacity_per_country_interp = capacity_per_country.copy()
    min_year = min(year_cols)
    max_year = max(year_cols)
    for year in range(min_year, max_year + 1, 1):
        if year not in capacity_per_country_interp.columns:
            capacity_per_country_interp.loc[:, int(year)] = np.nan
    capacity_per_country_interp = capacity_per_country_interp.sort_index(axis=1).interpolate(axis=1)

    # Calculate the gradient (ratio of growth between consecutive years) and drop irrelevant years
    interp_year_cols = [
        col for col in capacity_per_country_interp.columns if isinstance(col, int) and 1900 <= int(col) <= 3000
    ]  # Select only years
    year_cols_to_drop = [col for col in interp_year_cols if isinstance(col, int) and col <= base_yr]
    # TODO: Check div by zero
    capacity_gradient = capacity_per_country_interp.pct_change(axis=1).drop(columns=year_cols_to_drop)

    # Use the gradient to project the capacity from the base year onto [base_year +1, 2100]
    capacity_projection = (
        pd.merge(current_capacity, capacity_gradient.reset_index(), on="Country", how="left")
        .rename(columns={f"Capacity {tech} (MW)": base_yr})
        .set_index(["Country", "Scenario"])
    )
    projection_years = cast(list[int], sorted([col for col in capacity_projection.columns if isinstance(col, int)]))
    for i in range(1, len(projection_years)):
        year = projection_years[i]
        prev_year = projection_years[i - 1]
        capacity_projection[year] = capacity_projection[prev_year] * (1 + capacity_projection[year])

    return capacity_projection


def update_capex_using_learning_curve(capacity_0: float, capacity_t: float, capex_0: float, lr: float) -> float:
    """
    Update the capex based on the capacity change.
    Note: Capacity units are irrelevant, since they cancel out.
    Cases with capacity_0 == 0 are replaced by small epsilon to avoid division by 0.
    """
    b = np.log(1 - lr) / np.log(2)
    # TODO: Move epsilon to global vars
    epsilon = 1e-3
    capacity_0 = max(capacity_0, epsilon)
    capex_t = capex_0 * (capacity_t / capacity_0) ** b

    return capex_t


def project_capex_across_all_dims(
    capacity_projections: pd.DataFrame,
    costs: pd.DataFrame,
    learning_rates: pd.DataFrame,
    base_yr: int,
    tech: str,
) -> pd.DataFrame:
    """
    Projects the CAPEX for a given technology across all dimensions (countries, scenarios, learning rates, and years) using the learning curve.
    """
    logging.info(f"Projecting CAPEX for {tech}")
    # Initialize an empty dataframe
    countries = capacity_projections.index.get_level_values("Country").unique()
    ssp_scenarios = capacity_projections.index.get_level_values("Scenario").unique().sort_values()
    index = pd.MultiIndex.from_product(
        [countries, ssp_scenarios, learning_rates.index], names=["Country", "Scenario", "LR Scenario"]
    )
    years = capacity_projections.columns
    capex_projection = pd.DataFrame(index=index, columns=years)

    # Store the base year CAPEX data in the empty dataframe
    # TODO: Add split up between hardware and non-hardware capex if needed, currently using totals
    initial_capex = (
        costs[f"Total capex {tech} (USD/kW)"].to_frame().rename(columns={f"Total capex {tech} (USD/kW)": base_yr})
    )
    initial_capex = initial_capex.reindex(capex_projection.index, level="Country")
    capex_projection.loc[:, base_yr] = initial_capex.values

    # Project CAPEX using the learning curve
    for scenario in ssp_scenarios:
        for lr_scen in learning_rates.index:
            for country in countries:
                for i in range(1, len(years)):
                    year = years[i]
                    prev_year = years[i - 1]
                    capex_0_ = capex_projection.loc[pd.IndexSlice[country, scenario, lr_scen], prev_year]
                    capacity_0_ = capacity_projections.loc[(country, scenario), prev_year]
                    capacity_t_ = capacity_projections.loc[(country, scenario), year]
                    capex_t = update_capex_using_learning_curve(
                        capacity_0_, capacity_t_, capex_0_, LEARNING_RATES.loc[lr_scen][tech]
                    )
                    capex_projection.loc[pd.IndexSlice[country, scenario, lr_scen], year] = capex_t

    return capex_projection


def calculate_yearly_generated_electricity_in_lifetime(
    capacity_factor: float, lifetime: int, deterioration_rate: float
) -> list[float]:
    """
    Calculate the yearly generated electricity in MWh by 1 MW os installed capacity over its lifetime, taking into account its
    deterioration; 8760 = hours in a year (24 h/day x 365 days/year).
    Inputs:
        capacity_factor: ratio between total generated electricity and total electricity we could potentially generate
        if all installed capacity was running at its fullest (%).
        years: list of years in the lifetime of the technology.
        deterioration_rate: rate at which the capacity deteriorates with each year that passes by (%).
    """
    generated_electricity: list[float] = []
    # Initialize full capacity and actual capacity arrays
    full_capacity = [1.0] * lifetime
    actual_capacity = full_capacity.copy()

    # Calculate actual capacity with deterioration
    for i in range(1, lifetime):  # No deterioration in the first year
        actual_capacity[i] = actual_capacity[i - 1] * (1 - deterioration_rate)

    # Calculate generated electricity for each year
    for year in range(lifetime):
        # TODO: Move 8760 to global vars
        generated_electricity.append(actual_capacity[year] * capacity_factor * 8760)

    return generated_electricity


def calculate_lcoe(
    generated_electricity: list[float],
    fixed_opex_percentage: float,
    cost_of_capital: float,
    capex_0: float,
    capex_t: list[float] | None = None,
    curtailment: list[float] | None = None,
) -> float:
    """
    Calculate the Levelized Cost of Electricity (LCOE) in USD/MWh.
    lcoe = (capex_0 + sum_t(fixed_opex_percentage * capex_t / (1 + cost_of_capital)^t)) /
            sum_t((generated_electricity_t * (1 - curtailment_t)) / (1 + cost_of_capital)^t))
    Notes:
        1. sum_t from t = 1 to T, where T is the number of years in the lifetime of the technology. t=0 (investment time)
        corresponds to CAPEX_0 and generated electricity = 0.
        2. Can be calculated without absolute capacity values, since it cancels out anyways (set capacity to 1 MW and use
        CAPEX per MW).

    Inputs:
        generated_electricity: Electricity generated per timestep in MW.
        capex_0: Initial capital expenditure (CAPEX) in USD/MW.
        fixed_opex_percentage: Fixed OPEX as a percentage of CAPEX in %.
        cost_of_capital: Discount rate (r) in % (risk aversion, discount gets higher with time as uncertainty increases).
        curtailment: Curtailment rate per timestep in %.

    Output:
        lcoe: Levelized Cost of Electricity (LCOE) in USD/MWh.
    """
    # Simplification depending on cases
    if not capex_t:
        capex_t = [capex_0] * len(generated_electricity)
    if not curtailment:
        curtailment = [0] * len(generated_electricity)
    years = range(len(generated_electricity))
    discount_factors = [
        (1 + cost_of_capital) ** (t + 1) for t in years
    ]  # (t + 1) since the index starts at 0 and t starts at 1
    # Numerator
    x = [capex_t[t] / discount_factors[t] for t in years]
    opex_proxy = fixed_opex_percentage * sum(x)
    numerator = capex_0 + opex_proxy
    # Denominator
    sold_electricity = [generated_electricity[t] * (1 - curtailment[t]) for t in years]
    discounted_electricity = [sold_electricity[t] / discount_factors[t] for t in years]
    denominator = sum(discounted_electricity)
    if denominator == 0:
        raise ZeroDivisionError("Denominator (discounted electricity) is 0, cannot divide by 0.")

    return numerator / denominator


def get_lcoe_across_all_dims(
    capacity_fact: pd.Series | pd.DataFrame,
    capex_projection: pd.DataFrame,
    cost: pd.DataFrame,
    learning_rates: pd.DataFrame,
    lifetime: int,
    deterioration: float,
    tech: str,
) -> pd.DataFrame:
    """
    Calculate the LCOE for all combinations of countries, scenarios, and learning rates and store it in a dataframe with a
    multiindex for those dimensions.
    """
    logging.info(f"Calculating LCOE for {tech}")
    # Correct units
    capex_projection_usdmw = capex_projection * 1000  # from USD/kW to USD/MW
    capex_columns = cost.columns[cost.columns.str.contains("capex")]
    cost_usdmw = cost.copy()
    cost_usdmw[capex_columns] = cost[capex_columns] * 1000  # from USD/kW to USD/MW
    cost_usdmw = cost_usdmw.rename(columns={col: str(col)[:-8] + "(USD/MW)" for col in capex_columns})

    # Initialize an empty dataframe to store the LCOE results
    countries = capex_projection.index.get_level_values("Country").unique()
    ssp_scenarios = capex_projection.index.get_level_values("Scenario").unique().sort_values()
    investment_years = list(
        capex_projection.columns[:-lifetime]
    )  # Select only years for which the whole lifetime can be calculated
    index = pd.MultiIndex.from_product(
        [countries, ssp_scenarios, learning_rates.index], names=["Country", "Scenario", "LR Scenario"]
    )
    lcoe = pd.DataFrame(index=index, columns=[investment_years])

    # Calculate the LCOE for all combinations of countries, scenarios, and learning rates
    # TODO: Speed up by reducing times stuff gets calculated (e.g., discount factors for each year, calculte only once for all 70 years, read from there). Lambda fct instead of for loop?
    for scenario in ssp_scenarios:
        for lr_scen in learning_rates.index:
            for country in countries:
                for investment_year in investment_years:
                    # Select the time horizon for LCOE calculation (years). The investment year (t=0) corresponds to CAPEX_0
                    # and generated electricity = 0, which is why it is excluded from the time steps.
                    years = range(int(investment_year) + 1, int(investment_year) + lifetime + 1)
                    generated_electricity = calculate_yearly_generated_electricity_in_lifetime(
                        capacity_fact.loc[country],
                        lifetime,
                        deterioration,
                    )
                    lcoe_value = calculate_lcoe(
                        generated_electricity,
                        cost_usdmw.loc[country, f"Opex {tech} (%)"],
                        cost_usdmw.loc[country, "Cost of capital (%)"],
                        capex_projection_usdmw.loc[(country, scenario, lr_scen), investment_year],
                        [capex_projection_usdmw.loc[(country, scenario, lr_scen), year] for year in years],
                    )
                    lcoe.loc[(country, scenario, lr_scen), investment_year] = lcoe_value
    return lcoe


# ----------------------------------------------------- MAIN -----------------------------------------------------
def execute_calculate_cost_of_renewables(
    country_region: dict, country_code: dict, country_irena: dict, lr: pd.DataFrame, life: dict, det_rate: dict
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Main function to project the LCOE of renewables (solar PV and onshore wind) at country level until 2100 minus their lifetime.
    """

    # Select base year - year of the "current" CAPEX and capacity data
    base_year = 2022

    # Load data
    raw_solar_capex, raw_wind_capex, raw_cost_data = load_cost_data(base_yr=base_year)
    raw_capacity_factors, raw_hist_capacity, raw_solar_capacity_ssp, raw_wind_capacity_ssp = load_capacity_data(
        base_yr=base_year
    )

    # Preprocess data
    countries = pd.DataFrame(list(country_region.keys()), columns=["Country"])
    cost_per_country = preprocess_cost_data(
        raw_solar_capex, raw_wind_capex, raw_cost_data, countries, country_region, country_code
    )
    capacity_factors = preprocess_capacity_factors(raw_capacity_factors)
    solar_capacity_now, wind_capacity_now = preprocess_historical_renewable_capacity(
        raw_hist_capacity, countries, country_irena
    )
    solar_capacity = correct_ssp_projections_with_current_capacity(
        solar_capacity_now, raw_solar_capacity_ssp, countries, COUNTRY_TO_SSP_REGION_MAP, base_year, "solar"
    )
    wind_capacity = correct_ssp_projections_with_current_capacity(
        wind_capacity_now, raw_wind_capacity_ssp, countries, COUNTRY_TO_SSP_REGION_MAP, base_year, "wind"
    )

    # Project CAPEX based on the learning curve
    solar_capex_projection = project_capex_across_all_dims(solar_capacity, cost_per_country, lr, base_year, "solar")
    wind_capex_projection = project_capex_across_all_dims(wind_capacity, cost_per_country, lr, base_year, "wind")

    # Calculate LCOE for solar and wind
    solar_lcoe = get_lcoe_across_all_dims(
        capacity_factors["Capacity factor solar (%)"],
        solar_capex_projection,
        cost_per_country,
        lr,
        life["solar"],
        det_rate["solar"],
        "solar",
    )
    solar_lcoe.replace([np.inf, -np.inf], np.nan, inplace=True)

    wind_lcoe = get_lcoe_across_all_dims(
        capacity_factors["Capacity factor wind (%)"],
        wind_capex_projection,
        cost_per_country,
        lr,
        life["wind"],
        det_rate["wind"],
        "wind",
    )
    wind_lcoe.replace([np.inf, -np.inf], np.nan, inplace=True)

    return solar_lcoe, wind_lcoe


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    solar_lcoe_, wind_lcoe_ = execute_calculate_cost_of_renewables(
        COUNTRY_TO_REGION_MAP,
        COUNTRY_TO_CODE_MAP,
        COUNTRY_TO_IRENA_MAP,
        LEARNING_RATES,
        LIFETIME,
        YEARLY_DETERIORATION_RATE,
    )
    logging.info(f"Solar LCOE: \n{solar_lcoe_}")
    logging.info(f"Wind LCOE: \n{wind_lcoe_}")



from typing import Callable

from ..domain import events, commands, Year
from .unit_of_work import UnitOfWork
from ..domain.models import Environment
from steelo.utilities.global_variables import ANNOUNCED_STATUSES
from datetime import datetime


import logging

logger = logging.getLogger(__name__)


def close_furnace_group(cmd: commands.CloseFurnaceGroup, uow: UnitOfWork):
    with uow:
        plant = uow.plants.get(cmd.plant_id)
        plant.close_furnace_group(cmd.furnace_group_id)
        uow.commit()


def renovate_furnace_group(cmd: commands.RenovateFurnaceGroup, uow: UnitOfWork):
    with uow:
        plant = uow.plants.get(cmd.plant_id)
        plant.renovate_furnace_group(cmd.furnace_group_id)
        uow.commit()


def change_furnace_group_technology(cmd: commands.ChangeFurnaceGroupTechnology, uow: UnitOfWork):
    with uow:
        plant = uow.plants.get(cmd.plant_id)
        plant.change_furnace_group_technology(cmd.furnace_group_id, cmd.technology_name)
        uow.commit()


def add_furnace_group(cmd: commands.AddFurnaceGroup, uow: UnitOfWork):
    with uow:
        plant = uow.plants.get(cmd.plant_id)
        plant.add_furnace_group(cmd.new_furnace)
        plant.added_capacity += cmd.new_furnace.capacity
        plant.furnace_group_added(cmd.furnace_group_id, cmd.plant_id)
        uow.commit()


def update_cost_curve(_event: events.Event, uow: UnitOfWork, env: Environment):
    with uow:
        env.update_cost_curve(world_furnace_groups=[fg for plant in uow.plants.list() for fg in plant.furnace_groups])
        uow.commit()


def update_furnace_utilization_rates(event: events.SteelAllocationsCalculated, uow: UnitOfWork):
    allocation = event.allocations.allocations
    with uow:
        for plant in uow.plants.list():
            for fg in plant.furnace_groups:
                fg.utilization_rate = 0.0
        # furnace_groups = {fg.furnace_group_id: fg for plant in uow.plants.list() for fg in plant.furnace_groups}
        for (plant_obj, furnace_group, _), production in allocation.items():
            furnace_group.utilization_rate = production / furnace_group.capacity
        for plant in uow.plants.list():
            for fg in plant.furnace_groups:
                if fg.utilization_rate == 0.0:
                    uow.plants.get(plant.plant_id).close_furnace_group(fg.furnace_group_id)
        uow.commit()


def finalise_iteration(event: events.IterationOver, env: Environment, uow: UnitOfWork):
    env.year = Year(env.year + event.time_step_increment)
    with uow:
        for plant in uow.plants.list():
            plant.reset_capacity_changes()
            for fg in plant.furnace_groups:
                fg.lifetime.current = env.year
                if fg.lifetime.current >= fg.lifetime.time_frame.start and fg.status.lower() in ANNOUNCED_STATUSES:
                    fg.status = "Operating"
        uow.commit()
    logger.debug(f"finalising iteration. time: {datetime.now()}")


EVENT_HANDLERS: dict[type[events.Event], list[Callable]] = {
    events.FurnaceGroupClosed: [update_cost_curve],
    events.FurnaceGroupTechChanged: [update_cost_curve],
    events.FurnaceGroupRenovated: [update_cost_curve],
    events.FurnaceGroupAdded: [update_cost_curve],
    events.SteelAllocationsCalculated: [update_furnace_utilization_rates, update_cost_curve],
    events.IterationOver: [finalise_iteration],
}

COMMAND_HANDLERS: dict[type[commands.Command], Callable] = {
    commands.CloseFurnaceGroup: close_furnace_group,
    commands.RenovateFurnaceGroup: renovate_furnace_group,
    commands.ChangeFurnaceGroupTechnology: change_furnace_group_technology,
    commands.AddFurnaceGroup: add_furnace_group,
}



from typing import Any, NewType, Sequence


class _Auto:
    """
    Sentinel value indicating an automatic default will be used.
    """

    def __bool__(self):
        # Allow `Auto` to be used like `None` or `False` in boolean expressions
        return False


Auto: Any = _Auto()
Volumes = NewType("Volumes", int)  # Volumes of steel or iron (unit: ttpa)
Year = NewType("Year", int)
Years = Sequence[Year]



import logging

from datetime import date

from .calculate_costs import (
    calculate_cost_of_stranded_asset,
    calculate_lost_cash_flow,
    calculate_gross_cash_flow,
    BillsOfMaterial,
)
from .models import FurnaceGroup

# Configure logging at the start of your script
# Level can be DEBUG, INFO, WARNING, ERROR, CRITICAL
logging.basicConfig(
    level=logging.DEBUG,  # sets the minimum logging level
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


def stranding_asset_cost(
    opex: list[float] | list[int],
    total_debt_repayment: list[float] | list[int],
    price: int | float,
    expected_production: int | float,
    lifetime_remaining: int,
    cost_of_equity: float,
) -> float:
    """
    This is simply a wrapper function that takes agent level insights and computes the cost of stranding an asset given it's current situtation
    """
    remaining_debt = total_debt_repayment[-lifetime_remaining:]
    remaining_opex = opex[-lifetime_remaining:]

    gross_cash_flow = calculate_gross_cash_flow(remaining_opex, price, expected_production)
    lost_cf = calculate_lost_cash_flow(remaining_debt, gross_cash_flow)
    return calculate_cost_of_stranded_asset(lost_cf, cost_of_equity)


# def agent_decision(
#     # current_cost: float,
#     lifetime_remaining: int,
#     main_production_equipment: str,
#     UR: int,
#     expected_production: int,
#     cost_of_equity: float,
#     cosa: float,
#     new_asset_lifetime: int = 20,
#     high_production_threshold: float = 0.95,
#     low_production_threshold: float = 0.6,
#     tech_npv: dict = {"EAF": 1, "DRI-EAF": 1.1, "BF-BOF": 1.3},
# ) -> str:
#     """
#     Determines the cost-optimal decision for an asset agent, dynamically adjusting based on economic conditions.
#     Parameters:
#     current_cost (float): Current cost of production
#     lifetime_remaining (int): Remaining lifetime of the asset
#     main_production_equipment (str): Current production technology
#     expected_production (int): Expected production for the asset
#     r (float): Discount rate
#     price (float): Price of the product
#     OPEX (float): Operating expenses
#     remaining_capital_debt (float): the remaining debt still owed on the asset
#     new_asset_lifetime (int): Lifetime of the new asset
#     pain_point (float): Pain threshold for CoSA impact on decision
#     high_production_threshold (float): High production threshold
#     low_production_threshold (float): Low production threshold
#     """
#     # Calculate NPV for current and alternative technologies
#     # tech_npv = NPV_function(expected_production, prev_tech=main_production_equipment)
#     if UR < low_production_threshold:
#         return "close"

#     # Calculate Capital Recovery Factor (CRF)
#     crf = (cost_of_equity * (1 + cost_of_equity) ** new_asset_lifetime) / (
#         (1 + cost_of_equity) ** new_asset_lifetime - 1
#     )

#     # Determine Opportunity Cost with CoSA considered
#     opportunity_value = calculate_opportunity_cost(tech_npv, main_production_equipment, cosa, crf)
#     print(opportunity_value)
#     # Check if economic situation favors a decision before end of life
#     # Find the minimum value in the row
#     optimal_tech, init_npv = optimal_decision(opportunity_value)

#     if lifetime_remaining == 0:
#         # Conduct end-of-life decision
#         return end_of_life_decision(
#             tech_npv,
#             main_production_equipment,
#             expected_production,
#             high_production_threshold,
#             low_production_threshold,
#         )

#     else:
#         # Find the best option after accounting for CoSA

#         # Choose action based on expected production and adjusted opportunity cost
#         if main_production_equipment == optimal_tech:
#             return evaluate_high_or_medium_production(UR, high_production_threshold, low_production_threshold)
#         else:
#             decision = evaluate_high_or_medium_production(UR, high_production_threshold, low_production_threshold)
#             if decision == "expand":
#                 return f"switch to {optimal_tech} consider expansion"
#             else:
#                 return f"switch to {optimal_tech}"


def optimal_decision(cost_of_opportunity: dict[str, float]) -> dict[str, float]:
    """
    Determines the cost-optimal decision for an asset agent, dynamically adjusting based on economic conditions.
    Parameters:
    cost_of_opportunity (pd.DataFrame): Cost of opportunity for each technology
    main_production_equipment (str): Current production technology

    Returns:
    list: the optimal production equipments based on npv
    """

    return max(cost_of_opportunity.items(), key=lambda item: item[1])  # type: ignore


# def evaluate_high_or_medium_production(expected_production, high_threshold, low_threshold):
#     """
#     Evaluates expand/renovate/close based on production level
#     """
#     if expected_production >= high_threshold:
#         return "expand"
#     elif low_threshold <= expected_production < high_threshold:
#         return "renovate / do nothing"
#     else:
#         return "close"


def calculate_opportunity_cost(
    tech_npv: dict[str, float], main_production_equipment: str, cost_of_stranding_asset: float, crf: float
) -> dict[str, float]:
    """
    Calculates the opportunity cost, adding stranded asset costs to technologies other than the main one.
    """
    stranded_costs = crf * cost_of_stranding_asset
    for tech, npv in tech_npv.items():
        if tech != main_production_equipment:
            tech_npv[tech] -= stranded_costs
    return tech_npv


# FIXME: This function cannot possibly work as it is (missing plant_id)
# def decision_cycle(
#     furnace_group: FurnaceGroup, optimal_tech: str, init_npv: float, asset_lifetime_remaining: int, message_bus
# ) -> None:
#     """
#     Makes a decision based on opportunity cost at the end of asset life.
#     And adjusts the furnace group in question accordingly
#     """
#
#     if init_npv < 0:
#         logging.warning("all NPV <0! Renovating this furnace group would lead to financial ruin. We will close it.")
#         message_bus.handle(CloseFurnaceGroup(furnace_group.furnace_group_id))
#         return
#
#     cmd: Command
#     if furnace_group.technology.name == optimal_tech:
#         if asset_lifetime_remaining <= 0:
#             furnace_group.last_renovation_date = date.today()
#             logging.info(
#                 f"The asset is operating the most profitable technology {furnace_group.technology.name}. SO we renovate"
#             )
#             cmd = RenovateFurnaceGroup(furnace_group_id=furnace_group.furnace_group_id)
#             message_bus.handle(cmd)
#             return
#             # consider_expanding(furnace_group, unit_fopex=1, capacity_increment=furnace_group.capacity * 0.5, demand=300)
#         else:
#             logging.info(
#                 f"The asset is operating the most profitable technology {furnace_group.technology.name} and will therefore continue operation as is"
#             )
#             return
#
#     else:
#         logging.info(f"Switching furnace_group to optimal technology {optimal_tech}")
#         technology = Technology(
#             name=optimal_tech,
#             bill_of_materials=getattr(BillsOfMaterial, optimal_tech.replace("-", "_")),
#             product=furnace_group.technology.product,
#         )
#         cmd = ChangeFurnaceGroupTechnology(
#             furnace_group_id=furnace_group.furnace_group_id, technology_name=technology.name
#         )
#         message_bus.handle(cmd)


def asset_cosa(furnace_group: FurnaceGroup, asset_lifetime_remaining: int, cost_of_equity: float, price: int) -> float:
    remaining_opex = [int(furnace_group.opex)] * asset_lifetime_remaining
    remaining_debt = furnace_group.debt_repayment_per_year

    gross_cash_flow = calculate_gross_cash_flow(
        opex=remaining_opex, price=price, expected_production=int(furnace_group.production)
    )
    lost_cf = calculate_lost_cash_flow(remaining_debt, gross_cash_flow)
    return calculate_cost_of_stranded_asset(lost_cf, cost_of_equity)


def switch_to_optimal_technology(fg: FurnaceGroup, optimal_tech: str):
    fg.technology.name = optimal_tech
    fg.technology.bill_of_materials = getattr(BillsOfMaterial, optimal_tech.replace("-", "_"))
    fg.last_renovation_date = date.today()

    return fg


# def consider_expanding(
#     fg: FurnaceGroup,
#     unit_fopex: float,
#     capacity_increment: float,
#     demand: float,
#     environment: Environment,
#     risk_aversion_pct: float = 0.1,
# ) -> str:
#     capex = fg.technology.greenfield_capex

#     current_price = environment.extract_price_from_costcurve(demand=demand)
#     init_npv = _npv_flow_wrapper(fg, fg.technology.name, unit_fopex=unit_fopex, price=current_price, capex=capex)

#     production_cost_dict = dict(
#         production=(fg.utilization_rate - risk_aversion_pct) * fg.capacity + capacity_increment,
#         unit_cost_of_production=fg.unit_production_cost,
#     )
#     ### This bit is slightly difficult
#     # Trigger cost curve recalculation
#     new_price = environment._update_dict_and_extract_price(
#         new_entry=production_cost_dict, furnace_group_id=fg.furnace_group_id, demand=demand
#     )
#     ###

#     expand_npv = _npv_flow_wrapper(
#         fg, switch_technology=fg.technology.name, capex=capex, unit_fopex=unit_fopex, price=new_price
#     )

#     print(f"expansion impact on price: current price {current_price}, new price {new_price}")
#     print(f"expansion impact: current npv: {init_npv}, new npv if expanding {expand_npv}")
#     if expand_npv <= init_npv:
#         return "Do not expand"
#     else:
#         return "expand"


# def technology_switch(optimal_tech, product):
#     logging.info(f"Switching furnace_group to optimal technology {optimal_tech}")
#     technology = Technology(
#         name=optimal_tech,
#         bill_of_materials=getattr(BillsOfMaterial, optimal_tech.replace("-", "_")),
#         product=product,
#     )



from typing import Any

from pyomo.environ import (  # type: ignore
    ConcreteModel,
    Set,
    Param,
    Var,
    NonNegativeReals,
    Objective,
    Constraint,
    minimize,
    SolverFactory,
)

import networkx as nx  # type: ignore
import math
from enum import Enum
import time

from steelo.domain import Auto

# Constants
GRAPH_DATA_TOLERANCE = 3
SOURCE_NODE_TYPE = "source"
SINK_NODE_TYPE = "sink"
COST_DECIMAL_CORRECTION = 100


class PseudoNodes(Enum):
    PSEUDO_SOURCE_NAME = "Pseudo-Source"
    PSEUDO_SINK_NAME = "Pseudo-Sink"


class HitchcockProblem:
    def __init__(self, sources, sinks, allocation_costs, source_costs=None):
        """
        Initializes the Hitchcock Transportation Problem.

        Args:
        sources (dict): A dictionary with source nodes as keys and their supply amounts as values.
        sinks (dict): A dictionary with sink nodes as keys and their demand amounts as values.
        allocation_costs (dict of dict): A nested dictionary where allocation_costs[i][j] is the cost to transport from source i to sink j.

        Note: Unlike traditional Hitchcock Problems, this structure can also solve the problem if total demand != total supply,
        however only if solves as a min-cost max-flow problem, otherwise it leads to an LP infeasibility.
        """
        self.sources = sources
        self.sinks = sinks
        self.allocation_costs = allocation_costs
        self.source_costs = source_costs

    def display_problem(self):
        """Prints the supply, demand, and costs in a readable format."""
        print("Supply:")
        for s in self.sources:
            print(f" Source {s}: {self.sources[s]}")

        print("\nDemand:")
        for t in self.sinks:
            print(f" Sink {t}: {self.sinks[t]}")

        print("\nAllocation Costs:")
        for s in self.sources:
            for t in self.sinks:
                if t in self.allocation_costs[s]:
                    print(f" Cost from Source {s} to Sink {t}: {self.allocation_costs[s][t]}")

    def to_networkx_graph(self):
        """
        Creates a networkx graph from the Hitchcock problem instance.

        Returns:
        G (networkx.DiGraph): A directed graph where nodes represent sources and sinks,
                               and edges represent transportation routes with attributes for costs and capacities.
        """
        G = nx.DiGraph()

        # Add source and sink nodes with their supplies and demands as node attributes
        for source, supply in self.sources.items():
            G.add_node(source, supply=int(supply), demand=0, node_type=SOURCE_NODE_TYPE)
        for sink, demand in self.sinks.items():
            G.add_node(sink, demand=int(demand), supply=0, node_type=SINK_NODE_TYPE)

        # Add edges with costs and capacities as edge attributes
        for source in self.allocation_costs:
            for sink in self.allocation_costs[source]:
                assert sink in self.sinks, f"Sink {sink} not in the list of sinks."
                assert source in self.sources, f"Source {source} not in the list of sources."
                G.add_edge(
                    source,
                    sink,
                    cost=int(COST_DECIMAL_CORRECTION * self.allocation_costs[source][sink]),
                    capacity=1_000_000,
                )

        return G

    def turn_to_max_flow_min_cost_problem(self):
        """
        Augments a given graph from a Hitchcock Problem with a pseudo-source and pseudo-sink.

        Returns:
        networkx.DiGraph: The augmented graph with a pseudo-source and pseudo-sink as well as adapted edge costs and capacities.
        """
        G = self.to_networkx_graph()
        H = G.copy()

        # Add pseudo-source and pseudo-sink nodes
        pseudo_source = PseudoNodes.PSEUDO_SOURCE_NAME.value
        pseudo_sink = PseudoNodes.PSEUDO_SINK_NAME.value

        total_source_supply = round(
            sum(data["supply"] for node, data in G.nodes(data=True) if data["node_type"] == SOURCE_NODE_TYPE)
        )
        total_sink_demand = round(
            sum(
                data["demand"] if not math.isnan(data["demand"]) else 0
                for node, data in G.nodes(data=True)
                if data["node_type"] == SINK_NODE_TYPE
            )
        )

        H.add_node(pseudo_source, supply=int(total_source_supply), demand=0, node_type=pseudo_source)
        H.add_node(pseudo_sink, supply=0, demand=int(total_sink_demand), node_type=pseudo_sink)

        # Connect the pseudo-source to all sources
        for node, data in G.nodes(data=True):
            if data["node_type"] == SOURCE_NODE_TYPE:
                if self.source_costs is None or math.isinf(self.source_costs[node]):
                    # FIXME 0 cost for infinite costs - is this correct?
                    cost = 0
                else:
                    cost = (
                        int(COST_DECIMAL_CORRECTION * self.source_costs[node]) if self.source_costs is not None else 0
                    )
                H.add_edge(pseudo_source, node, capacity=int(data["supply"]), cost=cost)
                H.nodes[node]["supply"] = 0

        # Connect all sinks to the pseudo-sink
        for node, data in G.nodes(data=True):
            if data["node_type"] == SINK_NODE_TYPE:
                H.add_edge(node, pseudo_sink, capacity=int(data["demand"]), cost=0)
                H.nodes[node]["demand"] = 0

        return H

    @staticmethod
    def check_correctness(graph):
        """Checks the correctness of the graph structure."""
        for edge in graph.edges:
            if "capacity" in graph.edges[edge]:
                edge_cap = graph.edges[edge]["capacity"]
                if not isinstance(edge_cap, int):
                    print(f"Capacity of edge {edge} is not integral: {edge_cap}")
                    return False
                if edge_cap < 0:
                    print(f"Capacity of edge {edge} is negative: {edge_cap}")
                    return False

            if "cost" in graph.edges[edge]:
                edge_cost = graph.edges[edge]["cost"]
                if not isinstance(edge_cost, int):
                    print(f"Cost of edge {edge} is not integral: {edge_cost}")
                    return False
                if edge_cost < 0:
                    print(f"Cost of edge {edge} is negative: {edge_cost}")
                    return False

        for node in graph.nodes:
            node_demand = graph.nodes[node]["demand"]
            node_supply = graph.nodes[node]["supply"]
            if not isinstance(node_demand, int) or not isinstance(node_supply, int):
                print(f"Supply or demand of node {node} is not integral: {node_supply}, {node_demand}")
                return False
            if node_demand < 0 or node_supply < 0:
                print(f"Supply or demand of node {node} is negative: {node_supply}, {node_demand}")
                return False

        return True

    def solve_as_min_cost_flow(self):
        """Solves the problem as a minimum-cost flow problem."""
        H = self.turn_to_max_flow_min_cost_problem()
        if not self.check_correctness(H):
            print("Error: There's something wrong with the graph.")
            return None, None

        # Check for cycles
        if list(nx.simple_cycles(H)):
            print("Error: Graph contains cycles.")
            return None, None

        start_time = time.time()
        min_cost_flow = nx.max_flow_min_cost(
            H,
            PseudoNodes.PSEUDO_SOURCE_NAME.value,
            PseudoNodes.PSEUDO_SINK_NAME.value,
            capacity="capacity",
            weight="cost",
        )
        execution_time = time.time() - start_time
        min_cost = nx.cost_of_flow(H, min_cost_flow, weight="cost")

        print(f"Execution time: {execution_time:.2f} seconds")
        return min_cost, self.remove_pseudo_nodes(min_cost_flow)

    def remove_pseudo_nodes(self, min_cost_flow):
        # Remove arcs from and to the pseudo source and sink
        pseudo_source = PseudoNodes.PSEUDO_SOURCE_NAME.value
        pseudo_sink = PseudoNodes.PSEUDO_SINK_NAME.value

        for node in list(min_cost_flow[pseudo_source].keys()):
            del min_cost_flow[pseudo_source][node]
        for node in list(min_cost_flow.keys()):
            if pseudo_sink in min_cost_flow[node]:
                del min_cost_flow[node][pseudo_sink]
        return min_cost_flow

    def solve_as_lp(self, *, solver=Auto):
        """
        Solves the Hitchcock (transportation) problem using Pyomo.

        Returns:
            tuple: (total minimal cost, flow configuration)
        """
        # Create a Pyomo model
        model = ConcreteModel()

        # Define sets
        model.SOURCES = Set(initialize=sorted(self.sources.keys()))
        model.SINKS = Set(initialize=sorted(self.sinks.keys()))

        # Define parameters
        #    supply[s] = capacity of source s
        #    demand[t] = demand of sink t
        #    cost[s, t] = cost to ship from s to t (only if s->t exists in allocation_costs)
        def supply_init(model, s):
            return self.sources[s]

        model.supply = Param(model.SOURCES, initialize=supply_init)

        def demand_init(model, t):
            return self.sinks[t]

        model.demand = Param(model.SINKS, initialize=demand_init)

        # We need a cost for each (s, t). If some (s, t) isn't in self.allocation_costs[s],
        # you can set it to a large number or exclude it entirely. Here we assume the user
        # only includes valid s->t pairs in self.allocation_costs.
        def cost_init(model, s, t):
            return self.allocation_costs[s][t]

        model.cost = Param(model.SOURCES, model.SINKS, initialize=cost_init)

        # 4. Define variables:
        #    x[s, t] >= 0 is the flow from source s to sink t.
        model.x = Var(model.SOURCES, model.SINKS, domain=NonNegativeReals)

        # Objective: Minimize total cost = sum_{s in SOURCES, t in SINKS} cost[s,t] * x[s,t]
        def obj_rule(model):
            return sum(model.cost[s, t] * model.x[s, t] for s in model.SOURCES for t in model.SINKS)

        model.obj = Objective(rule=obj_rule, sense=minimize)

        # Constraints
        # Supply constraints: sum of flows from each source s <= supply[s]
        def supply_constraint_rule(model, s):
            return sum(model.x[s, t] for t in model.SINKS) <= model.supply[s]

        model.supply_constraint = Constraint(model.SOURCES, rule=supply_constraint_rule)

        # Demand constraints: sum of flows into each sink t = demand[t]
        def demand_constraint_rule(model, t):
            return sum(model.x[s, t] for s in model.SOURCES) == model.demand[t]

        model.demand_constraint = Constraint(model.SINKS, rule=demand_constraint_rule)

        if not solver:
            # Solve the model with a solver of your choice, e.g. "glpk" or "cbc"
            solver = SolverFactory("cbc")  # or "glpk", etc.

            # Check if solver is available
            if not solver.available():
                raise Exception("Solver not available. Please install 'cbc' or another compatible solver.")
        results = solver.solve(model, tee=False)

        # Check solver status
        if (results.solver.status != "ok") or (results.solver.termination_condition != "optimal"):
            raise Exception("No feasible (optimal) solution found.")

        # Extract solution
        total_cost = model.obj()
        flow: dict[Any, dict[Any, float]] = {}
        for s in model.SOURCES:
            flow[s] = {}
            for t in model.SINKS:
                flow[s][t] = model.x[s, t].value

        return total_cost, flow



from typing import Callable

import numpy as np
import geopandas as gpd  # type: ignore
import shapely.geometry as shp_geom
import osmnx as ox
from sklearn.preprocessing import MinMaxScaler  # type: ignore


class GeoSpatialAdapter:
    @staticmethod
    def get_country_boundary(country_name: str) -> gpd.GeoDataFrame:
        boundary = ox.geocode_to_gdf(country_name)
        return boundary.to_crs("EPSG:4326")

    @staticmethod
    def create_grid(boundary: gpd.GeoDataFrame, cell_size: float) -> list[dict[str, float]]:
        minx, miny, maxx, maxy = boundary.total_bounds

        # Use numpy.arange to handle floating-point cell sizes
        x_coords = np.arange(minx, maxx, cell_size)
        y_coords = np.arange(miny, maxy, cell_size)

        grid_polygons = [shp_geom.box(x, y, x + cell_size, y + cell_size) for x in x_coords for y in y_coords]
        grid = gpd.GeoDataFrame({"geometry": grid_polygons}, crs="EPSG:4326")

        # Ensure valid intersection
        clipped_grid = gpd.overlay(grid, boundary, how="intersection")
        clipped_grid = clipped_grid[clipped_grid.is_valid]  # Filter invalid geometries

        result = []
        for geometry in clipped_grid.geometry:
            if geometry.is_empty:  # Skip empty geometries
                continue
            bounds = geometry.bounds  # Extract numeric bounds directly
            result.append(
                {
                    "geometry": geometry,
                    "minx": float(bounds[0]),
                    "miny": float(bounds[1]),
                    "maxx": float(bounds[2]),
                    "maxy": float(bounds[3]),
                }
            )
        return result

    @staticmethod
    def compute_distances(
        grid: list[dict[str, float]], infrastructure: dict[str, gpd.GeoDataFrame]
    ) -> list[dict[str, float]]:
        for infra_type, gdf in infrastructure.items():
            # Re-project the infrastructure to a suitable CRS (e.g., EPSG:3857 for meters)
            projected_gdf = gdf.to_crs("EPSG:3857")

            for row in grid:
                # Re-project the grid point to match the projected CRS
                point = shp_geom.Point((row["minx"], row["miny"]))
                projected_point = gpd.GeoSeries([point], crs="EPSG:4326").to_crs("EPSG:3857").iloc[0]

                # Calculate distance in meters
                row[f"distance_to_{infra_type}_meters"] = projected_gdf.geometry.distance(projected_point).min()

        return grid

    @staticmethod
    def get_scaler() -> Callable[[list[list[float]]], list[list[float]]]:
        """
        Provide a scaling function using scikit-learn's MinMaxScaler.

        Returns:
            Callable: A scaling function that can be used in the domain.
        """
        scaler = MinMaxScaler()

        def scale(data: list[list[float]]) -> list[list[float]]:
            return scaler.fit_transform(data).tolist()

        return scale



import pandas as pd


def read_prep_data(path_to_data):
    data = pd.read_excel(path_to_data, sheet_name="Outputs for model")
    data = data[data.Side == "Input"].copy()
    data["Value"] = data.apply(
        lambda x: x["Value"] / 1000 if x["Unit"] == "kg/t product" else x["Value"], axis=1
    ).values
    data["Unit"] = data.apply(lambda x: "t/t product" if x["Unit"] == "kg/t product" else x["Unit"], axis=1).values
    data["Value"] = data.apply(
        lambda x: x["Value"] * 227.8 if "GJ" in x["Unit"] and x["Vector"] == "Electricity" else x["Value"], axis=1
    ).values
    data["Unit"] = data.apply(
        lambda x: "kWh/t product" if x["Unit"] == "GJ/t product" and x["Vector"] == "Electricity" else x["Unit"],
        axis=1,
    ).values

    return (
        data.fillna("null")
        .groupby(
            ["Business case", "Metallic charge", "Reductant", "Metric type", "Vector", "Type", "Unit"], dropna=False
        )["Value"]
        .sum()
        .reset_index(["Unit"])
        .to_dict("index")
        .copy()
    )


def main(path_to_data, path_to_output):
    output = read_prep_data(path_to_data)

    content = f"""
bom_data = {output}
"""
    with open(path_to_output, "w") as file:
        file.write(content)


if __name__ == "__main__":
    main(path_to_data="data/2024_12_13_New business cases_MVP.xlsx", path_to_output="src/steelo/domain/BOM.py")



from datetime import date

from .domain import Auto, Plant, Location, Technology, FurnaceGroup, ProductCategory, Volumes, Year, DemandCenter
from .domain.models import PointInTime, TimeFrame

technology_consumption_dict = {
    "BF-BOF": {
        "materials": {
            "Iron Ore": {"demand": 2, "unit_cost": {"Value": 1.5}},
            "Scrap": {"demand": 0.2, "unit_cost": {"Value": 3.0}},
        },
        "energy": {
            "Electricity": {"demand": -1.0, "unit_cost": {"Value": 0.75}},
            "Hydrogen": {"demand": 0.0, "unit_cost": {"Value": 2.5}},  # 5 times electricity
            "Coal": {"demand": 10.0, "unit_cost": {"Value": 0.5}},
            "Gas": {"demand": 5.0, "unit_cost": {"Value": 0.5}},
        },
    },
    "DRI-EAF": {
        "materials": {
            "Iron Ore": {"demand": 2, "unit_cost": {"Value": 1.5}},
            "Scrap": {"demand": 0.2, "unit_cost": {"Value": 3.0}},
        },
        "energy": {
            "Electricity": {"demand": 6.0, "unit_cost": {"Value": 0.75}},
            "Hydrogen": {"demand": 5.0, "unit_cost": {"Value": 2.5}},
            "Coal": {"demand": 0.0, "unit_cost": {"Value": 0.5}},
        },
    },
    "EAF": {
        "materials": {
            "Iron": {"demand": 0.2, "unit_cost": {"Value": 2.5}},
            "Scrap": {"demand": 1.2, "unit_cost": {"Value": 3.0}},
        },
        "energy": {
            "Electricity": {"demand": 6.0, "unit_cost": {"Value": 0.75}},
            "Hydrogen": {"demand": 0.0, "unit_cost": {"Value": 2.5}},
            "Coal": {"demand": 0.0, "unit_cost": {"Value": 0.5}},
        },
    },
}


capex_switch_dict = {
    "BF-BOF": {"EAF": 10, "DRI-EAF": 13, "BF-BOF": 4},
    "DRI-EAF": {"EAF": 5, "DRI-EAF": 5.5, "BF-BOF": 8},
    "EAF": {"EAF": 5, "DRI-EAF": 8, "BF-BOF": 8},
}


def get_furnace_group(
    *,
    fg_id: str = "fg_group_1",
    tech_name: str = "EAF",
    production: float = 45.0,
    utilization_rate: float = 0.7,
    lifetime: PointInTime = Auto,
    capacity: Volumes = Auto,
) -> FurnaceGroup:
    technology = Technology(
        name=tech_name,
        bill_of_materials=technology_consumption_dict[tech_name],
        product="Steel",
    )
    if not lifetime:
        lifetime = PointInTime(current=Year(2025), time_frame=TimeFrame(start=Year(2015), end=Year(2035)))
    if not capacity:
        capacity = Volumes(int(production / utilization_rate))
    furnace_group = FurnaceGroup(
        furnace_group_id=fg_id,
        capacity=capacity,
        status="Operating",
        last_renovation_date=date(2015, 5, 4),  # TODO @Beth
        technology=technology,
        historical_production={},
        utilization_rate=utilization_rate,
        lifetime=lifetime,
    )
    return furnace_group


def get_plant(
    *,
    plant_id: str = "plant_1",
    tech_name: str = "EAF",
    unit_production_cost: float = 70.0,
    production: float = 45.0,
    location: Location = Auto,
    furnace_groups: list[FurnaceGroup] = Auto,
) -> Plant:
    if not location:
        location = Location(iso3="DEU", country="Germany", region="Europe", lat=49.40768, lon=8.69079)
    if not furnace_groups:
        furnace_groups = [
            get_furnace_group(
                fg_id=f"{plant_id}_fg",
                tech_name=tech_name,
                production=production,
            )
        ]
    return Plant(
        plant_id=plant_id,
        location=location,
        furnace_groups=furnace_groups,
        power_source="unknown",
        soe_status="unknown",
        parent_gem_id="gem_id",
        workforce_size=10,
        certified=False,  # TODO @Beth
        category_steel_product={ProductCategory("Flat")},
    )


def get_demand_center(
    *,
    centre_id: str = "demand_center_1",
    gravity: Location = Auto,
    demand: dict[Year, Volumes] = Auto,
) -> DemandCenter:
    if not demand:
        demand = {Year(2025): Volumes(10), Year(2026): Volumes(20)}
    return DemandCenter(
        demand_center_id=centre_id,
        center_of_gravity=gravity,
        demand_by_year=demand,
    )


furnace_announcement_dates = [
    ("announced", 2028),
    ("Operating", 2024),
    ("Operating", 2024),
    ("announced", 2030),
    ("Operating", 2024),
    ("announced", 2031),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("announced", 2035),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("announced", 2030),
    ("Operating", 2024),
    ("announced", 2032),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("announced", 2031),
    ("construction", 2034),
    ("announced", 2032),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("announced", 2030),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("announced", 2031),
    ("Operating", 2024),
    ("Operating", 2024),
    ("announced", 2032),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
    ("Operating", 2024),
]



from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Dict, Any
import logging
import json
import os

from .service_layer.message_bus import MessageBus
from .service_layer.unit_of_work import UnitOfWork
from .economic_models import EconomicModel, PlantAgentsModel, AllocationModel
from .domain.models import Environment
from .domain.events import IterationOver
from .domain.datacollector import DataCollector
from .domain.calculate_costs import BillsOfMaterial
from .bootstrap import bootstrap
from .config import settings
from .adapters.repositories.in_memory_repository import import_plants_from_json
from .adapters.dataprocessing.preprocessing.raw_plant_data_processing import read_technology_lcop_data
from .adapters.dataprocessing.excel_reader import read_demand_centers_into_repository
from .utilities.global_variables import ACTIVE_STATUSES


class Simulation:
    """
    A class to run a simulation using an economic model. The simulation
    uses a MessageBus to handle events and commands and uses the strategy
    pattern to run the economic model.
    """

    def __init__(self, *, bus: MessageBus, economic_model: EconomicModel) -> None:
        self.bus = bus
        self.economic_model = economic_model

    def run_simulation(self) -> None:
        self.economic_model.run(self.bus)


@dataclass
class SimulationConfig:
    """Configuration for running a complete simulation."""

    plants_json_path: Path = settings.sample_plants_json_path
    technology_lcop_csv: Path = settings.technology_lcop_csv
    gravity_distances_csv: Path = settings.gravity_distances_csv
    start_year: int = 2025
    end_year: int = 2027
    output_file: Path = settings.intermediate_outputs_simulation_run
    demand_center_xlsx: Path = settings.demand_center_xlsx
    demand_sheet_name: str = "Steel_Demand_Chris Bataille"
    location_csv: Path = settings.location_csv
    log_level: int = logging.WARNING
    cost_of_x_csv: Path = settings.cost_of_x_csv

    def __repr__(self):
        return json.dumps(self.__dict__, indent=4, default=str)


class SimulationRunner:
    """
    A class that orchestrates the full simulation process.
    This can be used from both CLI and Django interfaces.
    """

    def __init__(self, config: SimulationConfig, do_setup: bool = True):
        self.config = config
        logging.getLogger().setLevel(config.log_level)

        # Results that will be populated after setup
        self.data_collector: Optional[DataCollector] = None
        self.bus: Optional[MessageBus] = None

        # Run setup during initialization to populate required attributes
        if do_setup:
            # FIXME: this should be handled in a better way
            self.setup()

    def setup(self):
        """Set up all necessary components for the simulation."""
        # Load plant data from JSON
        repository_json = import_plants_from_json(json_path=self.config.plants_json_path)
        technology_lcops = read_technology_lcop_data(self.config.technology_lcop_csv)

        # Set technology LCOP data for each plant
        for plant in repository_json.plants.list():
            plant.set_technology_lcops(technology_lcops)
            plant.calculate_average_steel_cost_and_capacity()

        # Read demand centers
        repository = read_demand_centers_into_repository(
            demand_excel_path=self.config.demand_center_xlsx,
            demand_sheet_name=self.config.demand_sheet_name,
            location_csv=self.config.location_csv,
            repository=repository_json,
            gravity_distances_csv=self.config.gravity_distances_csv,
        )

        # Set bill of materials for each furnace group
        for plant in repository_json.plants.list():
            for fg in plant.furnace_groups:
                fg.technology.bill_of_materials = getattr(BillsOfMaterial, fg.technology.name.replace("-", "_"))

        # Create environment and message bus
        env = Environment(cost_of_x_csv=self.config.cost_of_x_csv)
        uow = UnitOfWork(repository=repository)
        bus = bootstrap(uow=uow, env=env)

        # Initialize environment
        steel_furnaces = [
            fg
            for plant in bus.uow.plants.list()
            for fg in plant.furnace_groups
            if fg.technology.product in ["steel", "Steel"]
        ]
        bus.env.initiate_demand_dicts(repository.demand_centers.list())
        bus.env.calculate_demand()
        bus.env.generate_cost_curve(steel_furnaces)
        bus.env.update_regional_capacity(bus.uow.plants.list())

        # Create data collector
        data_collector = DataCollector(world_plants=bus.uow.plants.list(), env=bus.env)

        self.bus = bus
        self.data_collector = data_collector

    def print_capacity_stats(self):
        """Print statistics about plant capacities."""
        if not self.bus:
            raise RuntimeError("Bus is not initialized. Run setup() first.")

        capacities_operated = 0
        capacities_non_operated = 0

        for plant in self.bus.uow.plants.list():
            for fg in plant.furnace_groups:
                if fg.status.lower() in ACTIVE_STATUSES and fg.technology.product in ["steel", "Steel"]:
                    capacities_operated += fg.capacity
                elif fg.technology.product in ["steel", "Steel"]:
                    capacities_non_operated += fg.capacity

        print("Total Capacity of Operating Furnaces: ", capacities_operated)
        print("Total Capacity of Non-Operating Furnaces: ", capacities_non_operated)
        print("Steel Price:", self.bus.env.extract_price_from_costcurve(demand=self.bus.env.current_demand))

    def run(self) -> Dict[str, Any]:
        """Run the full simulation process and return results."""
        if not self.bus or not self.data_collector:
            raise RuntimeError("Bus or DataCollector is not initialized. Run setup() first.")

        # Initial statistics
        self.print_capacity_stats()

        # Run simulation for each year in the range
        for year in range(self.config.start_year, self.config.end_year):
            self.bus.env.calculate_demand()
            print("Demand: ", self.bus.env.current_demand)

            # Run the allocation and plant agent models
            Simulation(bus=self.bus, economic_model=AllocationModel()).run_simulation()
            Simulation(bus=self.bus, economic_model=PlantAgentsModel()).run_simulation()

            print("Year: ", year)

            # Collect data for this time step
            self.data_collector.collect()

            # Print statistics for this year
            self.print_capacity_stats()

            # Advance time step
            if (event := IterationOver) is not None:
                print("Event received: ", event)
                self.bus.handle(event(time_step_increment=1))

        # Prepare results dictionary
        results = {
            "price": self.data_collector.trace_price,
            "cost": self.data_collector.cost_breakdown,
            "capacity": self.data_collector.trace_capacity,
            "production": self.data_collector.trace_production,
        }

        # Save results to file if output_file is specified
        if self.config.output_file:
            output_file = self.config.output_file
            if os.path.exists(output_file):
                os.remove(output_file)
            with open(output_file, "w") as f:
                json.dump(results, f, indent=4)

        return results



import math
from pathlib import Path

import pandas as pd
from ...domain import (
    Location,
    Technology,
    FurnaceGroup,
    Plant,
    Volumes,
    ProductCategory,
    PointInTime,
    Year,
    TimeFrame,
    DemandCenter,
    PrimaryFeedstock,
    Auto,
)
from ..repositories import InMemoryRepository
from ..dataprocessing.preprocessing.raw_plant_data_processing import execute_preprocessing
from ..dataprocessing.preprocessing.iso3_finder import derive_iso3
from ...config import settings
from ...utilities import global_variables as gv

translate_country_names = {
    "Dem. Rep. of the Congo": "Congo DRC",
    "Hong Kong, China": "Hong Kong",
    "Ivory Coast": "Côte d'Ivoire",
    "Korea, North": "North Korea",
    "Korea, Republic of": "South Korea",
    "Macedonia": "North Macedonia",
    "Russia": "Russian Federation",
}

translate_technology_names = {
    "steel_eaf": "EAF",
    "steel_bof": "BOF",
    "iron_bf": "BF",
    "iron_dri": "DRI",
    "prep_pellet": "Prep Pellet",
    "prep_coke": "Prep Coke",
}


def read_location_from_row(row: dict) -> Location:
    """
    Create a Location object from a dictionary representing a data row.
    """
    return Location(
        lat=row["Latitude"],
        lon=row["Longitude"],
        country=row["Country"],
        region=row["Region"],
        iso3=derive_iso3(lat=row["Latitude"], lon=row["Longitude"]),
        # iso3="DEU",
    )


def read_technology_from_row(row: dict) -> Technology:
    """
    Create a Technology object from a dictionary representing a data row.
    Assigning the product based on the technology name:
      - "EAF" and "BOF" are associated with "Steel".
      - Other technologies are associated with "Iron".
    """
    tech = row["Technology"]
    if tech in ["EAF", "BOF"]:
        product = "Steel"
    else:
        product = "Iron"
    return Technology(
        name=row["Technology"],
        product=product,
    )


def read_pointintime_from_row(row: dict) -> PointInTime:
    """
    Create a PointInTime object from a row dict.

    Uses 'Last renovation year' if available, otherwise falls back to 'Start year'.
    The time frame spans from the base year to base year + 20.

    returns PointInTime object to the furnace group
    """
    if not pd.isna(row.get("Start year")):
        try:
            base_year = int(row["Last renovation year"])
        except (ValueError, TypeError):
            base_year = int(row["Start year"])
        return PointInTime(
            current=Year(gv.SIMULATION_START_YEAR),
            time_frame=TimeFrame(start=Year(base_year), end=Year(base_year + 20)),
        )
    else:
        return PointInTime(
            current=Year(gv.SIMULATION_START_YEAR),
            time_frame=TimeFrame(start=Year(0), end=Year(0)),
        )


def read_furnace_group_from_row(row: dict, iterator: int) -> FurnaceGroup:
    """
    Generate a FurnaceGroup object from a dictionary representing a data row from the GEM data
    This function reads the technology, capacity, and other relevant fields to construct a FurnaceGroup.
    The capacity column is determined based on the product type:
      - "Nominal steel capacity (ttpa)" for steel-based technologies.
      - "Nominal iron capacity (ttpa)" for iron-based technologies.
    """

    technology = read_technology_from_row(row)
    if technology.product == "Steel":
        cap_col = "Nominal steel capacity (ttpa)"
    else:
        cap_col = "Nominal iron capacity (ttpa)"

    try:
        capacity = Volumes(int(row[cap_col]))
    except ValueError:
        capacity = Volumes(0)

    lifetime = read_pointintime_from_row(row)

    return FurnaceGroup(
        furnace_group_id=f"{row['Plant ID']}_{iterator}",
        capacity=capacity,
        status=row["Capacity operating status"],
        lifetime=lifetime,
        last_renovation_date=pd.to_datetime(row["Last renovation year"]),
        technology=technology,
        historical_production={},
        utilization_rate=0.7,
    )


def has_location(row: dict) -> bool:
    """
    Check whether a data row contains valid location information.
    This function verifies that the row dictionary has both "Latitude" and "Longitude" keys with valid

    """
    lat, lon = row.get("Latitude"), row.get("Longitude")
    has_float_latitude = isinstance(lat, float) and not math.isnan(lat)
    has_float_longitude = isinstance(lon, float) and not math.isnan(lon)
    return has_float_latitude and has_float_longitude


def preprocessed_rows_to_domain_objects(
    preprocessed_rows: list[dict], gravity_distances_csv: Path = Auto
) -> dict[str, Plant]:
    """
    Convert a list of preprocessed data rows into domain Plant objects.

    Processes each row to create associated Location, FurnaceGroup, and Plant objects.
    Rows without valid location data are skipped. Additionally, gravity distance data is integrated into each location.
    """
    plants: dict[str, Plant] = {}
    if not gravity_distances_csv:
        gravity_distances_csv = settings.gravity_distances_csv
    gravity_df = pd.read_csv(gravity_distances_csv)
    for num, row in enumerate(preprocessed_rows):
        if not has_location(row):
            # If there's no location -> skip  FIXME maybe do something more sophisticated?
            continue
        location = read_location_from_row(row)
        read_gravity_distances_into_locations([location], gravity_df)

        furnace_group = read_furnace_group_from_row(row, num)
        # repository.furnace_groups.add(furnace_group)
        if (plant := plants.get(row["Plant ID"])) is not None:
            # plant already exists -> just add the furnace group and move on
            plant.furnace_groups.append(furnace_group)
            continue
        else:
            try:
                workforce_size = int(str(row["Workforce size"]))
            except ValueError:
                workforce_size = -1
            try:
                product_categories = {ProductCategory(str(row["Category steel product"]))}
            except ValueError:
                product_categories = set()
            plant = Plant(
                plant_id=str(row["Plant ID"]),
                location=location,
                furnace_groups=[furnace_group],
                power_source=str(row["Power source"]),
                soe_status=str(row["SOE Status"]),
                parent_gem_id=str(row["Parent GEM ID"]),
                workforce_size=workforce_size,
                certified=False,  # TODO @Beth
                category_steel_product=product_categories,
            )
            plants[plant.plant_id] = plant
    return plants


def read_gravity_distances_into_locations(locations: list[Location], gravity_df: pd.DataFrame):
    """
    Update a list of Location objects with gravity distance data.

    For each location, the function extracts gravity distance values from the provided DataFrame and assigns
    these distances to the location's 'distance_to_other_iso3' attribute.

    """
    for location in locations:
        gravity_this_location = gravity_df[gravity_df["iso3_o"] == location.iso3]
        gravity_dict = gravity_this_location.set_index("iso3_d")["dist"].to_dict()
        assert gravity_dict is not None
        location.distance_to_other_iso3 = gravity_dict


def read_plants_into_repository(gravity_distances_csv: Path = Auto) -> InMemoryRepository:
    """
    Load plant and furnace group data into in-memory repository.

    Processes preprocessed plant data to create Plant objects and their associated FurnaceGroup objects,
    and adds them to an InMemoryRepository.
    """
    repository = InMemoryRepository()

    # add plants
    _plant_df, furnace_group_df = execute_preprocessing()
    plants = preprocessed_rows_to_domain_objects(
        furnace_group_df.to_dict(orient="records"), gravity_distances_csv=gravity_distances_csv
    )
    repository.plants.add_list(plants.values())

    # add furnace groups
    furnace_groups = [fg for plant in plants.values() for fg in plant.furnace_groups]
    repository.furnace_groups.add_list(furnace_groups)
    return repository


def read_dynamic_business_cases_into_repository(
    dynamic_business_cases_excel_path: str, excel_sheet: str, repository: InMemoryRepository
):
    """
    Read dynamic business case data from an Excel file and organize it into a dictionary.

    Processes the specified Excel sheet to create PrimaryFeedstock objects based on business case information,
    constraints, material requirements, and energy requirements. Feedstocks are grouped by their technology.

    """
    dynamic_business_cases_df = pd.read_excel(dynamic_business_cases_excel_path, sheet_name=excel_sheet)
    read_feedstocks: dict[str, PrimaryFeedstock] = {}
    for _, row in dynamic_business_cases_df.iterrows():
        side = row["Side"]
        # only reading in inputs for now - as of now no motive for outputs
        if side == "Output":
            continue
        technology = translate_technology_names[row["Business case"]]
        metallic_charge = row["Metallic charge"]
        primary_feedstock_name = f"{technology}_{metallic_charge}"
        if primary_feedstock_name in read_feedstocks:
            primary_feedstock = read_feedstocks[primary_feedstock_name]
        else:
            primary_feedstock = PrimaryFeedstock(metallic_charge=metallic_charge, technology=technology)
        primary_feedstock.metallic_charge = metallic_charge
        primary_feedstock.technology = technology
        row_type = row["Metric type"]
        if row_type == "Constraint":
            constraint_type = row["Type"]
            assert isinstance(constraint_type, str), f"Expected str, got {type(constraint_type)}"
            value = row["Value"]
            primary_feedstock.add_share_constraint(constraint_type, value)
        if row_type == "Materials":
            material = row["Vector"]
            value = row["Value"]
            unit = row["Unit"]
            if material == metallic_charge:
                assert unit in [
                    "t/t product",
                    "t/t pellet",
                ], f"Unit '{unit}' not supported for primary material requirement."
                primary_feedstock.required_quantity_per_ton_of_product = value
            else:
                if unit in ["t/t product", "t/t pellet", "t/t coke"]:
                    converted_value = value
                elif unit == "kg/t product":
                    converted_value = value / 1000
                else:
                    raise ValueError(f"Unit '{unit}' not supported for secondary material requirement.")
                primary_feedstock.add_secondary_feedstock(material, converted_value)
        if row_type == "Energy":
            vector = row["Vector"]
            value = row["Value"]
            unit = row["Unit"]
            if unit.startswith("kWh/t"):
                converted_value = value
            elif unit.startswith("GJ/t"):
                converted_value = value * 0.0036
            elif unit == "kg/t product":
                # ugh, I don't know...
                converted_value = value
            else:
                print(f"Unit '{unit}' not supported for energy requirement.")
            primary_feedstock.add_energy_requirement(vector, converted_value)
        read_feedstocks[primary_feedstock_name] = primary_feedstock

    technology_feedstocks: dict[str, list[PrimaryFeedstock]] = {}
    for primary_feedstock in read_feedstocks.values():
        if primary_feedstock.technology not in technology_feedstocks:
            technology_feedstocks[primary_feedstock.technology] = []
        technology_feedstocks[primary_feedstock.technology].append(primary_feedstock)

    return technology_feedstocks


def read_demand_centers_into_repository(
    demand_excel_path: str,
    demand_sheet_name: str,
    location_csv: str,
    repository: InMemoryRepository = Auto,
    gravity_distances_csv: Path = Auto,
) -> InMemoryRepository:
    """
    Read demand center data from Excel and CSV files and add them into an in-memory repository.

    Performs the following steps:
      1. Reads gravity distance data from a CSV file.
      2. Loads and filters demand data from the specified Excel sheet.
      3. Translates country names using a predefined mapping.
      4. Reads location data from a CSV file and updates each Location with gravity distances.
      5. Constructs DemandCenter objects by matching demand data with locations.
      6. Adds the DemandCenter objects to the repository.
    """
    if not gravity_distances_csv:
        gravity_distances_csv = settings.gravity_distances_csv
    gravity_df = pd.read_csv(gravity_distances_csv)
    demand_df = pd.read_excel(demand_excel_path, sheet_name=demand_sheet_name)

    # TODO: What to do with demand from "other" areas -> might be fixed once we get actual demand data
    demand_df = demand_df[~demand_df["Country"].str.startswith("Other")]
    demand_df = demand_df[~demand_df["Country"].str.startswith("TOTAL")]
    demand_df["Country"] = demand_df["Country"].apply(lambda x: translate_country_names.get(x, x))
    location_df = pd.read_csv(location_csv)
    locations = []
    for _, row in location_df.iterrows():
        location = Location(
            lat=row["latitude"],
            lon=row["longitude"],
            country=row["COUNTRY"],
            region="Unknown",
            # FIXME 5000 km is a bit much, no?
            iso3=derive_iso3(lat=float(row["latitude"]), lon=float(row["longitude"]), max_distance_km=5000),
        )
        locations.append(location)

    # locations = [location for location in locations if location.iso3 in list(gravity_df.iso3_o.unique())]
    read_gravity_distances_into_locations(locations, gravity_df)

    demand_centers = []
    for _, row in demand_df.iterrows():
        try:
            demand_location = next(loc for loc in locations if loc.country == row["Country"])
        except StopIteration:
            raise ValueError(f"Location not found for country: {row['Country']}")
        demand_location.region = row["Region"]
        demand_by_year = {}
        for col in demand_df.columns:
            try:
                year = Year(int(col))
                demand_by_year[year] = Volumes(int(row[col]))
            except ValueError:
                continue
        demand_center = DemandCenter(
            demand_center_id=demand_location.country,
            center_of_gravity=demand_location,
            demand_by_year=demand_by_year,
        )
        demand_centers.append(demand_center)
    if not repository:
        repository = InMemoryRepository()
    repository.demand_centers.add_list(demand_centers)
    return repository



from typing import Protocol, runtime_checkable, Iterable

from ...domain import Plant, FurnaceGroup, DemandCenter


@runtime_checkable
class PlantRepository(Protocol):
    seen: set[Plant]

    def add(self, plant: Plant) -> None:
        """Add a steel plant to the repository."""
        ...

    def add_list(self, plants: Iterable[Plant]) -> None:
        """Add an iterable of steel plants to the repository."""
        ...

    def get(self, plant_id) -> Plant:
        """Get a steel plant from the repository by ID."""
        ...

    def list(self) -> list[Plant]:
        """Get a list of all steel plants."""
        ...

    def get_by_furnace_group_id(self, furnace_group_id: str) -> Plant:
        """Get a steel plant from the repository by furnace group ID."""
        ...


@runtime_checkable
class FurnaceGroupRepository(Protocol):
    def add(self, furnace_group: FurnaceGroup) -> None:
        """Add a furnace group to the repository."""
        ...

    def add_list(self, plants: Iterable[FurnaceGroup]) -> None:
        """Add an iterable of furnace groups to the repository."""
        ...

    def get(self, furnace_group) -> FurnaceGroup:
        """Get a furnace group from the repository by ID."""
        ...

    def list(self) -> list[FurnaceGroup]:
        """Get a list of all furnace groups."""
        ...


@runtime_checkable
class DemandCenterRepository(Protocol):
    def add(self, demand_center: DemandCenter) -> None:
        """Add a demand center to the repository."""
        ...

    def add_list(self, demand_center: Iterable[DemandCenter]) -> None:
        """Add an iterable of demand center to the repository."""
        ...

    def get(self, demand_center) -> DemandCenter:
        """Get a demand center from the repository by ID."""
        ...

    def list(self) -> list[DemandCenter]:
        """Get a list of all demand center."""
        ...


@runtime_checkable
class Repository(Protocol):
    plants: PlantRepository
    furnace_groups: FurnaceGroupRepository
    demand_centers: DemandCenterRepository



import pandas as pd

# GEM data years
STEEL_PLANT_GEM_DATA_YEAR = 2024
PRODUCTION_GEM_DATA_YEARS = range(2019, 2023)

# Learning rates for solar and wind energy
# Based on previous Systemiq project
LEARNING_RATES = pd.DataFrame(
    {
        "solar": [0.334, 0.234, 0.134],
        "wind": [0.226, 0.146, 0.066],
    },
    index=["low cost", "average cost", "high cost"],
)

# Lifetime of solar PV and on-shore wind turbines
LIFETIME = {
    "solar": 30,  # IRENA https://energycentral.com/system/files/ece/nodes/451407/irena_end_of_life_management_solar_pv.pdf p. 11; other sources say 25 years
    "wind": 25,  # IRENA https://www.irena.org/-/media/Files/IRENA/Agency/Publication/2017/Jun/IRENA_Leveraging_for_Onshore_Wind_Executive_Summary_2017.pdf p.20; other sources say 25 years
    "steel": 20,
}

# Deterioration rate the energy systems over their lifetime
# TODO: Look for better sources; currently using a rough estimate by Rafal
YEARLY_DETERIORATION_RATE = {
    "solar": 0.005,  # 0.5% per year
    "wind": 0.01,  # 1% per year
}


ENERGY_EMISSION_FACTORS = {
    "Thermal coal": 95 * 36,  # 95 kgCO2/GJ -> conversion to kgCO2/MWh
    "BF gas": 260 * 36,  # 260 kgCO2/GJ -> conversion to kgCO2/MWh
    "COG": 44 * 36,  # 44 kgCO2/GJ -> conversion to kgCO2/MWh
    "BOF gas": 192 * 36,  # 192 kgCO2/GJ -> conversion to kgCO2/MWh
    "Natural gas": 55 * 36,  # 55 kgCO2/GJ -> conversion to kgCO2/MWh
    "Plastic waste": 111 * 36,  # 111 kgCO2/GJ -> conversion to kgCO2/MWh
    "Electricity": 375,  # 375 gCO2/kWh
}

ACTIVE_STATUSES = ["operating", "operating pre-retirement"]
ANNOUNCED_STATUSES = ["announced", "construction"]

CAPACITY_LIMIT = 0.95

SIMULATION_START_YEAR = 2025



import pandas as pd
import geopandas as gpd  # type: ignore
from ....config import settings


# Function to Merge Data with GeoDataFrame
def merge_geographical_data(data_df: pd.DataFrame, country_column: str) -> pd.DataFrame:
    """
    Merge geographical data with statistical data based on ISO3 country codes.

    Parameters:
    data_df : pd.DataFrame
        Statistical data with country codes and values.
    geo_df : gpd.GeoDataFrame
        GeoDataFrame with country geometries.
    country_column : str
        Column in data_df with country codes matching 'iso_a3' in geo_df.

    Returns:
    gpd.GeoDataFrame
        Merged GeoDataFrame.
    """
    geo_df = gpd.read_file(
        settings.geo_boundaries_url,
    )
    if "iso_a3" not in geo_df.columns:
        raise ValueError("The GeoDataFrame does not contain the required 'iso_a3' column.")

    # Check if the country_column in data_df contains valid ISO3 codes
    if not data_df[country_column].str.match(r"^[A-Z]{3}$").all():
        raise ValueError(f"The column '{country_column}' does not contain valid ISO3 country codes.")

    return data_df.merge(geo_df, right_on="iso_a3", left_on=country_column)



import inspect

from .domain import Auto
from .service_layer import handlers, UnitOfWork, MessageBus
from .domain.models import Environment


def inject_dependencies(handler, dependencies):
    params = inspect.signature(handler).parameters
    deps = {name: dependency for name, dependency in dependencies.items() if name in params}
    return lambda message: handler(message, **deps)


def bootstrap(uow: UnitOfWork = Auto, env: Environment = Auto) -> MessageBus:
    """Creates a new MessageBus instance with all dependencies injected."""
    if not uow:
        uow = UnitOfWork()
    if not env:
        env = Environment()

    dependencies = {"uow": uow, "env": env}
    injected_event_handlers = {
        event_type: [inject_dependencies(handler, dependencies) for handler in event_handlers]
        for event_type, event_handlers in handlers.EVENT_HANDLERS.items()
    }
    injected_command_handlers = {
        command_type: inject_dependencies(handler, dependencies)
        for command_type, handler in handlers.COMMAND_HANDLERS.items()
    }
    return MessageBus(
        uow=uow, env=env, event_handlers=injected_event_handlers, command_handlers=injected_command_handlers
    )



from dataclasses import dataclass

from typing import TYPE_CHECKING

# from steelo.domain.datacollector import DataCollector

if TYPE_CHECKING:
    from .models import SteelAllocations


class Event:
    """Base class for all events."""


@dataclass
class FurnaceGroupClosed(Event):
    """Furnace group closed."""

    furnace_group_id: str


@dataclass
class FurnaceGroupTechChanged(Event):
    """Furnace group technology was changed."""

    furnace_group_id: str


@dataclass
class FurnaceGroupRenovated(Event):
    """Furnace group renovated."""

    furnace_group_id: str


@dataclass
class FurnaceGroupAdded(Event):
    """Furnace group renovated."""

    plant_id: str
    furnace_group_id: str


@dataclass
class SteelAllocationsCalculated(Event):
    """Steel allocations calculated."""

    allocations: "SteelAllocations"


@dataclass
class IterationOver(Event):
    """This timestep is finalised."""

    time_step_increment: int
    # dc: DataCollector



from collections import defaultdict
from typing import Any

from .unit_of_work import UnitOfWork
from .message_bus import MessageBus
from ..adapters.repositories import Repository

__all__ = ["MessageBus", "UnitOfWork"]


def get_markers(repository: Repository):
    plants = repository.plants.list()
    markers = []
    for plant in plants:
        marker = {
            "location": (plant.location.lat, plant.location.lon),
            "popup": plant.plant_id,
        }
        markers.append(marker)
    return markers


# Function to calculate capacity by technology and region
def calculate_capacity_by_technology_and_region(
    repository: Repository, product: str, furnace_statuses: list[str]
) -> defaultdict[Any, defaultdict[Any, int]]:
    plants = repository.plants.list()
    capacity_summary: defaultdict[Any, defaultdict[Any, int]] = defaultdict(lambda: defaultdict(int))

    for plant in plants:
        region = plant.location.region
        for furnace_group in plant.furnace_groups:
            if furnace_group.technology.product != product or furnace_group.status not in furnace_statuses:
                continue
            tech_name = furnace_group.technology.name
            capacity_summary[tech_name][region] += furnace_group.capacity

    return capacity_summary



import logging
from typing import TypedDict

from .dbom import derive_unnested_feedstock_bom, extract_steel, BusinessCaseArchitypes

from .constants import Auto


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

positive_infinity = float("inf")


def get_mvp_bom(technology: str):
    """
    Extract the bill of materials for the given business case archetype (no allocation data).

    Parameters:
        technology: The technology identifier used to look up the business case archetype.

    Returns:
        The extracted steel portion of the bill of materials.
    """
    business_case = BusinessCaseArchitypes.get_architype(technology)
    bom = derive_unnested_feedstock_bom(business_case)
    return extract_steel(bom)


class BillsOfMaterial:
    """
    Container for bills of materials for different technologies (cold-start mode without allocation data).

    NOTE: This class will become obsolete once trade allocation is used to distribute the feedstock
    among the different products.
    """

    BF_BOF = get_mvp_bom("BF-BOF")
    DRI_EAF = get_mvp_bom("DRI-EAF")
    EAF = get_mvp_bom("EAF")
    BF = get_mvp_bom("BF-BOF")  # TODO: update to BF-specific BOM
    BOF = get_mvp_bom("BF-BOF")  # TODO: update to BOF-specific BOM
    DRI = get_mvp_bom("DRI-EAF")  # TODO: update to DRI-specific BOM
    other = get_mvp_bom("BF-BOF")  # TODO: update to correct BOM for 'other'
    unknown = get_mvp_bom("BF-BOF")  # TODO: update to correct BOM for 'unknown'
    OHF = get_mvp_bom("BF-BOF")  # TODO: update to OHF-specific BOM


def calculate_variable_opex(materials_cost_data: dict, energy_cost_data: dict) -> int:
    """
    Calculate the total variable operating expenditure (OPEX).

    Parameters:
        materials_cost_data: Dictionary of material cost data with keys 'demand' and 'unit_cost' (with subkey "Value").
        energy_cost_data: Dictionary of energy cost data with keys 'demand' and 'unit_cost' (with subkey "Value").

    Returns:
        The total variable OPEX.
    """
    if not materials_cost_data:
        raise ValueError("materials_cost_data is empty.")

    material_cost = sum(data["demand"] * data["unit_cost"]["Value"] for data in materials_cost_data.values())
    energy_cost = sum(data["demand"] * data["unit_cost"]["Value"] for data in energy_cost_data.values())

    return material_cost + energy_cost


def calculate_fixed_opex(greenfield_capex: float, opex_capex_ratio: float = 0.1) -> float:
    """
    Calculate the fixed operating expenditure (O&M) as a fraction of the greenfield CAPEX.

    Parameters:
        greenfield_capex: The CAPEX related to building a new plant.
        opex_capex_ratio: The ratio of CAPEX allocated to fixed OPEX. Defaults to 0.1.

    Returns:
        The fixed OPEX.
    """
    return greenfield_capex * opex_capex_ratio


class CostData(TypedDict):
    demand: float
    unit_cost: dict[str, float]  # or a more specific TypedDict if you know more


BillOfMaterialsDict = dict[str, dict[str, CostData]]


def calculate_opex(bill_of_materials: BillOfMaterialsDict, fixed_opex: float, production: float) -> float:
    """
    Calculate the total operating expenditure (OPEX) per unit produced.

    Parameters:
        bill_of_materials: The bill of materials for producing one unit.
            Expected to have keys "materials" and "energy" with associated cost data.
        fixed_opex: The fixed operational expenditure.
        production: The production volume.

    Returns:
        The per-unit OPEX (fixed OPEX per unit plus variable OPEX per unit).
    """
    materials_cost = bill_of_materials["materials"]
    material_cost = sum(data["demand"] * data["unit_cost"]["Value"] for data in materials_cost.values())
    energy_cost_data = bill_of_materials["energy"]
    energy_cost = sum(data["demand"] * data["unit_cost"]["Value"] for data in energy_cost_data.values())
    variable_opex = material_cost + energy_cost
    return fixed_opex / production + variable_opex if production > 0 else positive_infinity  # to avoid division by zero


def calculate_debt_repayment(
    total_investment: float,
    equity_share: float,
    lifetime: int,
    cost_of_debt: float,
    lifetime_remaining: int | None = None,
) -> list[float]:
    """
    Calculate the yearly debt repayment schedule based on the investment parameters.

    Parameters:
        total_investment: The total investment cost.
        equity_share: The fraction of the investment financed by equity.
        lifetime: The full repayment lifetime in years.
        cost_of_debt: The annual interest rate on debt.
        lifetime_remaining: The remaining years for debt repayment. Defaults to lifetime.

    Returns:
        A list of yearly repayment amounts.
    """
    if lifetime_remaining is None:
        lifetime_remaining = lifetime

    debt = total_investment * (1 - equity_share)

    if debt == 0:
        total_repayment_list = [0.0] * lifetime
    else:
        capital_repayment_list = [debt / lifetime] * lifetime_remaining
        total_repayment_list = []
        for capital_repayment in capital_repayment_list:
            remaining_debt = debt - capital_repayment
            interest_repayment = ((debt + remaining_debt) / 2) * cost_of_debt
            total_repayment_list.append(interest_repayment + capital_repayment)
            debt = remaining_debt

    return total_repayment_list[-lifetime_remaining:]


def calculate_current_debt_repayment(
    total_investment: float,
    lifetime_expired: bool,
    lifetime_years: int,
    years_elapsed: int,
    cost_of_debt: float,
    equity_share: float,
) -> float:
    """
    Calculate the total debt repayment (principal plus interest) for the current year.

    Parameters:
        total_investment: The total investment cost.
        lifetime_expired: Flag indicating if the repayment period is over.
        lifetime_years: The total repayment lifetime in years.
        years_elapsed: The number of years elapsed since operation start.
        cost_of_debt: The annual interest rate on debt.
        equity_share: The fraction of the investment financed by equity.

    Returns:
        float: The debt repayment amount for the current year.
    """
    debt = total_investment * (1 - equity_share)

    if debt == 0 or lifetime_expired:
        return 0.0

    yearly_principal = debt / lifetime_years
    remaining_debt = debt - (yearly_principal * (years_elapsed - 1))
    avg_debt = (remaining_debt + (remaining_debt - yearly_principal)) / 2

    return yearly_principal + (avg_debt * cost_of_debt)


def calculate_unit_production_cost(opex: float, current_debt_repayment: float, production: float) -> float:
    """
    Calculate the unit production cost by adding OPEX and debt repayment costs per unit.

    Parameters:
        opex: The operating expenditure.
        current_debt_repayment: The debt repayment cost for the current period.
        production: The production volume.

    Returns:
        float: The cost per unit of production.
    """
    return opex + current_debt_repayment / production if production > 0 else positive_infinity


def calculate_gross_cash_flow_per_step(price: float, opex: float, expected_production: float) -> float:
    """
    Calculate the gross cash flow for a single time step.

    Parameters:
        price: The price per unit.
        opex: The operating expenditure per unit.
        expected_production: The production volume for the time step.

    Returns:
        float: The gross cash flow for the step.
    """
    return (price - opex) * expected_production


def calculate_gross_cash_flow(
    opex: list[float] | list[int], price: float | int, expected_production: float | int
) -> list[float]:
    """
    Calculate the gross cash flow over multiple time steps.

    Parameters:
        opex: List of operating expenditures per period.
        price: The price per unit.
        expected_production: The production volume per period.

    Returns:
        list[float]: A list of gross cash flows for each period.
    """
    return [calculate_gross_cash_flow_per_step(price, opex_year, expected_production) for opex_year in opex]


def calculate_net_cash_flow(total_debt_repayment: list[float], gross_cash_flow: list[float]) -> list[float]:
    """
    Calculate the net cash flow by subtracting debt repayment from gross cash flow for each period.

    Parameters:
        total_debt_repayment: List of debt repayment amounts per period.
        gross_cash_flow: List of gross cash flows per period.

    Returns:
        list[float]: A list of net cash flow values.
    """
    if len(total_debt_repayment) != len(gross_cash_flow):
        raise ValueError("The lengths of total_debt_repayment and gross_cash_flow must be the same.")

    return [gross - debt for gross, debt in zip(gross_cash_flow, total_debt_repayment)]


def calculate_lost_cash_flow(
    total_debt_repayment: list[float] | list[int], gross_cash_flow: list[float] | list[int]
) -> list[float]:
    """
    Calculate the lost cash flow by adding debt repayment to gross cash flow for each period.

    Parameters:
        total_debt_repayment: List of debt repayment amounts per period.
        gross_cash_flow: List of gross cash flows per period.

    Returns:
        list[float]: A list of lost cash flow values.
    """
    if len(total_debt_repayment) != len(gross_cash_flow):
        raise ValueError("The lengths of total_debt_repayment and gross_cash_flow must be the same.")

    return [gross + debt for gross, debt in zip(gross_cash_flow, total_debt_repayment)]


def calculate_cost_of_stranded_asset(lost_cash_flow: list[float], cost_of_debt: float) -> float:
    """
    Calculate the net present value (NPV) of losses due to stranding an asset.

    Parameters:
        lost_cash_flow: Future lost cash flows per period.
        cost_of_debt: The discount rate (cost of debt).

    Returns:
        float: The NPV of the stranded asset cost.
    """
    npv_loss = 0.0
    for t, cash in enumerate(lost_cash_flow, start=1):
        discount_factor = (1 + cost_of_debt) ** t
        npv_loss += cash / discount_factor
    return npv_loss


def calculate_npv_costs(
    net_cash_flow: list[float],
    cost_of_debt: float,
    equity_share: float,
    fixed_opex: float,
    total_investment: float = 0,
) -> float:
    """
    Calculate the net present value (NPV) of a series of cash flows.

    Parameters:
        net_cash_flow: List of future net cash flows per period.
        cost_of_debt: The discount rate (cost of debt).
        equity_share: The fraction of the investment financed by equity.
        fixed_opex: The fixed operating expenditure.
        total_investment: The total investment cost. Defaults to 0.

    Returns:
        float: The calculated NPV.
    """
    npv = -(total_investment * equity_share + fixed_opex)
    for t, cash in enumerate(net_cash_flow, start=1):
        npv += cash / ((1 + cost_of_debt) ** t)
    return npv


def calculate_levelised_cost(
    expected_production: list[float],
    opex: list[float],
    cost_of_debt: float,
    total_investment: float,
    total_repayment: list[float],
    equity_share: float,
    fixed_opex: float,
) -> float:
    """
    Calculate the levelised cost of steel or iron.
    The levelised cost of x is determined by discounting all costs and dividing by the production.

    Parameters:
        expected_production: Expected production per period.
        opex: Operating expenditures per period.
        cost_of_debt: Discount rate (cost of debt).
        total_investment: Total upfront investment cost.
        total_repayment: Debt repayments per period.
        equity_share: Fraction of the investment financed by equity.
        fixed_opex: Fixed operating expenditure at time zero.

    Raises:
        ValueError: If input lists have mismatched lengths or if discounted production is zero.
    """
    if not (len(expected_production) == len(opex) == len(total_repayment)):
        raise ValueError("Input lists 'expected_production', 'opex', and 'total_repayment' must have the same length.")

    initial_investment = total_investment * equity_share + fixed_opex
    discounted_cost = -initial_investment

    for t, (opex_year, repayment) in enumerate(zip(opex, total_repayment), start=1):
        discounted_cost -= (opex_year + repayment) / ((1 + cost_of_debt) ** t)

    discounted_production = sum(prod / ((1 + cost_of_debt) ** t) for t, prod in enumerate(expected_production, start=1))

    if discounted_production == 0:
        raise ValueError("Discounted production is zero, cannot compute levelised cost.")

    levelised_cost = discounted_cost / discounted_production
    return -levelised_cost


def calculate_npv_full(
    technology: str,
    capex: float,
    capacity: float,
    fixed_opex: float,
    expected_utilisation_rate: float,
    price: float,
    lifetime: float = 20,
    cost_of_debt: float = 0.05,
    equity_share: float = 0.2,
    bill_of_materials: dict[str, dict[str, dict[str, float]]] = Auto,
) -> float:
    """
    Calculate the net present value (NPV) for a technology.

    Parameters:
        technology: The technology identifier.
        capex: The capital expenditure per unit capacity.
        capacity: The capacity of the furnace group.
        fixed_opex: Fixed operating expenditure.
        expected_utilisation_rate: Expected utilisation rate.
        price: The market price per unit product.
        lifetime: The project lifetime in years. Defaults to 20.
        cost_of_debt: The cost of debt (interest rate). Defaults to 0.05.
        equity_share: The fraction of the investment financed by equity. Defaults to 0.2.
        bill_of_materials: Bill of materials data. If not provided, it is derived from BillsOfMaterial.

    Returns:
        The NPV for the technology.
    """
    if not bill_of_materials:
        bill_of_materials = getattr(BillsOfMaterial, technology.replace("-", "_"))

    variable_opex = calculate_variable_opex(
        bill_of_materials["materials"],
        bill_of_materials["energy"],
    )
    total_opex = fixed_opex + variable_opex
    total_investment = capacity * capex

    debt_repayment = calculate_debt_repayment(
        total_investment=total_investment, equity_share=equity_share, lifetime=int(lifetime), cost_of_debt=cost_of_debt
    )

    opex_list = [total_opex] * len(debt_repayment)
    production = expected_utilisation_rate * capacity
    gross_cash_flow = calculate_gross_cash_flow(opex=opex_list, price=price, expected_production=float(production))
    net_cash_flow = calculate_net_cash_flow(total_debt_repayment=debt_repayment, gross_cash_flow=gross_cash_flow)

    return calculate_npv_costs(
        net_cash_flow=net_cash_flow,
        cost_of_debt=cost_of_debt,
        equity_share=equity_share,
        fixed_opex=fixed_opex,
        total_investment=total_investment,
    )


def stranding_asset_cost(
    debt_repayment_per_year: list[float],
    opex: list[float],
    remaining_time: int,
    price: float,
    expected_production: float,
    cost_of_debt: float,
) -> float:
    """
    Calculate the cost associated with stranding an asset.

    Parameters:
        debt_repayment_per_year: Yearly debt repayment amounts.
        opex: Yearly operating expenditures.
        remaining_time: Remaining time (in years) for the asset.
        price: Price per unit product.
        expected_production: Production volume per period.
        cost_of_debt: Annual cost of debt (interest rate).

    Returns:
        The net present value of losses due to the stranded asset.
    """
    remaining_debt = debt_repayment_per_year[-remaining_time:]
    remaining_opex = opex[-remaining_time:]
    gross_cash_flow = calculate_gross_cash_flow(remaining_opex, price, expected_production)
    lost_cash_flow = calculate_lost_cash_flow(remaining_debt, gross_cash_flow)
    return calculate_cost_of_stranded_asset(lost_cash_flow, cost_of_debt)


def calculate_technogy_npv(
    capex_dict: dict[str, float],
    capacity: float,
    fixed_opex: float,
    expected_utilisation_rate: float,
    price: float,
    lifetime: int = 20,
    cost_of_debt: float = 0.05,
    equity_share: float = 0.2,
) -> dict[str, float]:
    """
    Calculate the net present value (NPV) for multiple technologies based on their CAPEX and other parameters.

    Parameters:
        capex_dict: Mapping of technology identifiers to CAPEX values.
        capacity: The capacity of the furnace group.
        fixed_opex: Fixed operating expenditure.
        expected_utilisation_rate: Expected utilisation rate.
        price: Market price per unit.
        lifetime: Lifetime in years. Defaults to 20.
        cost_of_debt: Cost of debt (interest rate). Defaults to 0.05.
        equity_share: Fraction of the investment financed by equity. Defaults to 0.2.

    Returns:
        A dictionary mapping technology identifiers to their NPV.
    """
    return {
        technology: calculate_npv_full(
            technology,
            capex,
            capacity,
            fixed_opex,
            expected_utilisation_rate,
            price,
            lifetime,
            cost_of_debt,
            equity_share,
        )
        for technology, capex in capex_dict.items()
    }.copy()


def derive_best_technology_switch(
    current_technology: str,
    cost_of_debt: float = 0.05,
    equity_share: float = 0.2,
    lifetime: int = 20,
    expected_utilisation_rate: float = 0.7,
    capacity: float = 1000,
    price: float = 100,
    mvp: bool = True,
    bill_of_materials: dict[str, dict[str, dict[str, float]]] = Auto,
    cosa=None,
) -> tuple[float, str]:
    """
    Calculate the net present value (NPV) for switching or renovating technologies and determine the best alternative.

    Parameters:
        current_technology: The current technology identifier.
        cost_of_debt: The cost of debt (interest rate). Defaults to 0.05.
        equity_share: The equity share. Defaults to 0.2.
        lifetime: The technology lifetime in years. Defaults to 20.
        expected_utilisation_rate: Expected utilisation rate. Defaults to 0.7.
        capacity: Capacity of the furnace group. Defaults to 1000.
        price: Market price per unit. Defaults to 100.
        mvp: Flag for MVP mode. Defaults to True.
        bill_of_materials: Bill of materials data. Defaults to Auto.
        cosa: Cost of stranding the current to subtract from non-current technologies.

    Returns:
        A tuple containing the best NPV and its corresponding technology identifier.
    """
    from steelo.domain.Capex import TECH_SWITCH_CAPEX, FULL_GREENFIELD_CAPEX

    npv_dict = {}
    if mvp:
        for tech, capex in TECH_SWITCH_CAPEX[current_technology].items():
            if tech in ["BOF", "DRI-EAF", "EAF"]:
                npv_dict[tech] = calculate_npv_full(
                    technology=tech,
                    capex=capex,
                    capacity=capacity,
                    fixed_opex=calculate_fixed_opex(FULL_GREENFIELD_CAPEX[tech]),
                    expected_utilisation_rate=expected_utilisation_rate,
                    price=price,
                    lifetime=lifetime,
                    cost_of_debt=cost_of_debt,
                    equity_share=equity_share,
                )
                if tech != current_technology and cosa is not None:
                    npv_dict[tech] -= cosa
    else:
        for tech, capex in TECH_SWITCH_CAPEX[current_technology].items():
            npv_dict[tech] = calculate_npv_full(
                technology=tech,
                capex=capex,
                capacity=capacity,
                fixed_opex=calculate_fixed_opex(FULL_GREENFIELD_CAPEX[tech]),
                expected_utilisation_rate=expected_utilisation_rate,
                price=price,
                lifetime=lifetime,
                cost_of_debt=cost_of_debt,
                equity_share=equity_share,
                bill_of_materials=bill_of_materials,
            )
            if tech != current_technology and cosa is not None:
                npv_dict[tech] -= cosa

    # Find tech with maximum NPV value
    best_tech = max(npv_dict, key=lambda k: npv_dict[k] if npv_dict[k] is not None else float("-inf"))
    best_value = npv_dict[best_tech]
    return best_value, best_tech


def calculate_capex_reduction_rate(
    capacity_zero: float,
    capacity_current: float,
    learning_coeff: float = -0.043943347587597055,
) -> float:
    """
    Calculate the CAPEX reduction rate based on learning effects.

    Parameters:
        capacity_zero: The initial capacity.
        capacity_current: The current capacity.
        learning_coeff: The learning coefficient. Defaults to -0.043943347587597055.

    Returns:
        The CAPEX reduction rate.
    """
    return (capacity_current / capacity_zero) ** learning_coeff


def calculate_debt_report(
    total_investment: float,
    lifetime_expired: bool,
    lifetime_years: int,
    years_elapsed: int,
    cost_of_debt: float,
    equity_share: float,
) -> tuple[float, float]:
    """
    Calculate the debt components (yearly principal and interest) for reporting purposes.

    Parameters:
        total_investment (float): The total investment cost.
        lifetime_expired (bool): Flag indicating if the debt repayment lifetime has expired.
        lifetime_years (int): The total number of repayment years.
        years_elapsed (int): The number of years that have elapsed.
        cost_of_debt (float): The annual interest rate on debt.
        equity_share (float): The fraction of the investment financed by equity.

    Returns:
        A tuple of (yearly_principal, interest_repayment) for the current year.
    """
    debt = total_investment * (1 - equity_share)

    if debt == 0 or lifetime_expired:
        return 0.0, 0.0

    yearly_principal = debt / lifetime_years
    remaining_debt = debt - (yearly_principal * (years_elapsed - 1))
    avg_debt = (remaining_debt + (remaining_debt - yearly_principal)) / 2

    return yearly_principal, (avg_debt * cost_of_debt)



import httpx
import osm2geojson  # type: ignore
import geopandas as gpd  # type: ignore


def fetch_and_convert_to_gdf(overpass_query, country):
    overpass_url = "http://overpass-api.de/api/interpreter"
    response = httpx.get(overpass_url, params={"data": overpass_query})
    response.raise_for_status()
    data = response.json()
    geojson = osm2geojson.json2geojson(data)
    gdf = gpd.GeoDataFrame.from_features(geojson["features"], crs="EPSG:4326")
    gdf["country"] = country
    return gdf


def extract_rail(country):
    overpass_query = f"""
    [out:json];
    area["name:en"={country}]->.searchArea;
    way["railway"="rail"]["usage"="main"](area.searchArea);
    out geom;
    """
    return fetch_and_convert_to_gdf(overpass_query, country)


def extract_highways(country):
    overpass_query = f"""
    [out:json];
    area["name:en"={country}]->.searchArea;
    way[highway~"^(motorway|trunk)$"](area.searchArea);
    out geom;
    """
    return fetch_and_convert_to_gdf(overpass_query, country)


def extract_power_lines(country):
    overpass_query = f"""
    [out:json][timeout:25];
    area["name:en"="{country}"]->.searchArea;
    (way["power"="line"](area.searchArea););
    out geom;
    """
    return fetch_and_convert_to_gdf(overpass_query, country)
